import argparse

def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('true', 't', '1'):
        return True
    elif v.lower() in ('false', 'f', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


import logging

from Bio import Phylo

from .base import Tree

class PrintTree(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        
        # remove branch lengths if specified by user
        # otherwise, print ascii tree
        if self.remove:
            for internode in tree.get_nonterminals():
                internode.branch_length = None
            for leaf in tree.get_terminals():
                leaf.branch_length = None

        try:
            Phylo.draw_ascii(tree)
        except BrokenPipeError:
            pass
    
    def process_args(self, args):
        return dict(tree_file_path=args.tree, remove=args.remove)


import getopt
import logging
import math
import os.path
import statistics as stat
import sys

from Bio import Phylo
import numpy as np


from .base import Tree

from ...helpers.stats_summary import calculate_summary_statistics_from_arr, print_summary_statistics

class InternalBranchStats(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        internal_branch_lengths, stats = self.calculate_internal_branch_stats(tree)

        if self.verbose:
            try:
                for internal_branch_length in internal_branch_lengths:
                    print(round(internal_branch_length, 4))
            except BrokenPipeError:
                pass
        else:
            print_summary_statistics(stats)

    def process_args(self, args):
        return dict(tree_file_path=args.tree, verbose=args.verbose)

    def get_internal_branch_lengths(self, tree) -> list:
        """
        loop through tree and get all internal branch lengths
        """
        internal_branch_lengths = []
        for internal_branch in tree.get_nonterminals():
            if internal_branch.branch_length != None:
                internal_branch_lengths.append(internal_branch.branch_length)

        return internal_branch_lengths

    def check_tree_has_branch_lengths(self, internal_branch_lengths:list) -> None:
        """
        if tree has no branch lengths, exit
        """
        if len(internal_branch_lengths) == 0:
            print("Calculating internal branch statistics requires a phylogeny with branch lengths.")
            sys.exit()


    def calculate_internal_branch_stats(self, tree):
        # save internal branch lengths to internal_branch_lengths
        internal_branch_lengths = self.get_internal_branch_lengths(tree)
        
        # If the phylogeny had no branch lengths, inform user and quit
        self.check_tree_has_branch_lengths(internal_branch_lengths)
        
        # calculate summary stats
        stats = calculate_summary_statistics_from_arr(internal_branch_lengths)

        return internal_branch_lengths, stats


import logging
import sys

from Bio import Phylo

from .base import Tree

class RenameTreeTips(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()

        # save idmap to a dictionary
        idmap = self.read_id_map(self.idmap)
        
        tree = self.replace_tip_names(tree, idmap)
        
        self.write_tree_file(tree, self.output_file_path)
    
    def process_args(self, args):
        tree_file_path = args.tree

        if args.output is None:
            output_file_path = f"{tree_file_path}.renamed"
        else:
            output_file_path = f"{args.output}"

        return dict(
            tree_file_path=tree_file_path, 
            idmap=args.idmap,
            output_file_path=output_file_path
            )

    def read_id_map(self, idmap: str) -> dict:
        """
        read two column file to dictionary
        """
        idmap={}
        try:
            with open(self.idmap) as identifiers:
                for line in identifiers:
                    (key, val) = line.split()
                    idmap[key] = val
        except FileNotFoundError:
            try:
                print(f"{self.idmap} corresponds to no such file.")
                print("Please check file name and pathing")
                sys.exit()
            except BrokenPipeError:
                pass

        return idmap

    def replace_tip_names(self, tree: Tree, idmap: dict) -> Tree:
        """
        if a tip name is in the idmap, replace it
        """
        # for tips with a name as a key in the idmap,
        # replace that tip name with the value in the
        # idmap
        for term in tree.get_terminals():
            if term.name in idmap:
                term.name = idmap[term.name]
        
        return tree




from collections import Counter
import itertools
import sys

from Bio import SeqIO

from .base import Alignment

class SumOfPairsScore(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # create biopython object of sequences for query and 
        # reference alignments
        query_records = SeqIO.to_dict(SeqIO.parse(self.fasta, "fasta"))
        reference_records = SeqIO.to_dict(SeqIO.parse(self.reference, "fasta"))

        # get all record pairs
        record_id_pairs = self.get_record_ids(reference_records)

        # calculate how many matches there are and how many total pairs there are
        number_of_matches, number_of_total_pairs = self.determine_number_of_matches_and_total_pairs(
            record_id_pairs,
            reference_records,
            query_records
        )
        
        # print res
        print(round(number_of_matches/number_of_total_pairs, 4))

    def process_args(self, args) -> dict:
        return dict(
            fasta=args.fasta,
            reference=args.reference
        )

    def get_record_ids(
        self,
        reference_records: dict
    ) -> list:
        # loop through record names and save each to 
        record_ids = []
        for entry_name in reference_records:
            record_ids.append(entry_name)
        # create all pairwise combinations
        record_id_pairs = list(itertools.combinations(record_ids, 2))
        return record_id_pairs

    def determine_number_of_matches_and_total_pairs(
        self,
        record_id_pairs: list,
        reference_records: dict,
        query_records: dict
    ):
        number_of_matches = 0
        number_of_total_pairs = 0
        # loop through each pair
        for record_pair in record_id_pairs:
            first_in_pair = record_pair[0]
            second_in_pair = record_pair[1]
            
            pairs_in_reference = []
            pairs_in_query = []
            # for each pair, loop through the length of the alignment and get sequence at each site
            for i in range(0, len(reference_records[first_in_pair].seq)):
                pairs_in_reference.append(reference_records[first_in_pair].seq[i] + reference_records[second_in_pair].seq[i])
            for i in range(0, len(query_records[first_in_pair].seq)):
                pairs_in_query.append(query_records[first_in_pair].seq[i] + query_records[second_in_pair].seq[i])

            # count the number of matches and total pairs
            matches = list((Counter(pairs_in_reference) & Counter(pairs_in_query)).elements())
            number_of_matches+=len(matches)
            number_of_total_pairs+=len(pairs_in_reference)

        return number_of_matches, number_of_total_pairs








import getopt
import logging
import math
import os.path
import statistics as stat
import sys

from Bio import Phylo
import numpy as np

from .base import Tree
from ...helpers.stats_summary import calculate_summary_statistics_from_arr, print_summary_statistics

class BipartitionSupportStats(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        bs_vals = self.get_bipartition_support_vals(tree) 

        if self.verbose:
            try:
                for bs_val in bs_vals:
                    print(bs_val)
            except BrokenPipeError:
                pass
        else:
            stats = calculate_summary_statistics_from_arr(bs_vals) 
            print_summary_statistics(stats)

    def process_args(self, args):
        return dict(tree_file_path=args.tree, verbose=args.verbose)

    def get_bipartition_support_vals(self, tree):
        # initialize list to hold bootstrap values
        bs_vals = []

        # populate bs_vals with bootstrap values
        for terminal in tree.get_nonterminals():
            # only include if a bootstrap value is present
            if terminal.confidence != None:
                bs_vals.append(terminal.confidence)
        return bs_vals


import logging

from Bio import Phylo

from .base import Tree

class TipLabels(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        try:
            for leaf in tree.get_terminals():
                print(leaf.name)
        except BrokenPipeError:
            pass
    
    def process_args(self, args):
        return dict(tree_file_path=args.tree)


import statistics as stat
import sys

import numpy as np


def calculate_summary_statistics_from_arr(
    arr    
):
    """
    calcuate summary statistics for an input list
    """
    stats = dict(
        mean               = stat.mean(arr),
        median             = stat.median(arr),
        twenty_fifth       = np.percentile(arr, 25),
        seventy_fifth      = np.percentile(arr, 75),
        minimum            = np.min(arr),
        maximum            = np.max(arr),
        standard_deviation = stat.stdev(arr),
        variance           = stat.variance(arr)
    )

    return stats

def calculate_summary_statistics_from_dict(
    dat: dict    
):
    """
    calcuate summary statistics for a dictionary
    """
    stats = dict(
        mean=stat.mean([*dat.values()]),
        median=stat.median([*dat.values()]),
        twenty_fifth=np.percentile([*dat.values()], 25),
        seventy_fifth=np.percentile([*dat.values()], 75),
        minimum=np.min([*dat.values()]),
        maximum=np.max([*dat.values()]),
        standard_deviation=stat.stdev([*dat.values()]),
        variance=stat.variance([*dat.values()])
    )

    return stats

def print_summary_statistics(
    stats: list
):
    """
    """
    try:
        print(f"mean: {round(stats['mean'], 4)}")
        print(f"median: {round(stats['median'], 4)}")
        print(f"25th percentile: {round(stats['twenty_fifth'], 4)}")
        print(f"75th percentile: {round(stats['seventy_fifth'], 4)}")
        print(f"minimum: {round(stats['minimum'], 4)}")
        print(f"maximum: {round(stats['maximum'], 4)}")
        print(f"standard deviation: {round(stats['standard_deviation'], 4)}")
        print(f"variance: {round(stats['variance'], 4)}")
    except BrokenPipeError:
        pass



from .bipartition_support_stats import BipartitionSupportStats
from .branch_length_multiplier import BranchLengthMultiplier
from .covarying_evolutionary_rates import CovaryingEvolutionaryRates
from .dvmc import DVMC
from .evolutionary_rate import EvolutionaryRate
from .internal_branch_stats import InternalBranchStats
from .internode_labeler import InternodeLabeler
from .last_common_ancestor_subtree import LastCommonAncestorSubtree
from .lb_score import LBScore
from .total_tree_length import TotalTreeLength
from .patristic_distances import PatristicDistances
from .polytomy_test import PolytomyTest
from .print_tree import PrintTree
from .prune_tree import PruneTree
from .rename_tree_tips import RenameTreeTips
from .rf_distance import RobinsonFouldsDistance
from .saturation import Saturation
from .spurious_sequence import SpuriousSequence
from .tip_labels import TipLabels
from .treeness import Treeness
from .treeness_over_rcv import TreenessOverRCV
from .collapse_branches import CollapseBranches


import getopt
import logging
import os.path
import sys

from Bio import Phylo

from .base import Tree

class BranchLengthMultiplier(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        tree_mod = self.multiply_branch_lengths_by_factor(tree, self.factor)
        self.write_tree_file(tree_mod, self.output_file_path)

    def process_args(self, args):
        if args.output is None:
            output_file_path = f"{args.tree}.factor_{args.factor}.tre"
        else:
            output_file_path = f"{args.output}"

        return dict(
            tree_file_path=args.tree,
            factor=args.factor,
            output_file_path=output_file_path
        )

    def multiply_branch_lengths_by_factor(self, tree, factor):
        for internode in tree.get_nonterminals():
            if internode.branch_length != None:
                internode.branch_length = internode.branch_length*factor
        for term in tree.get_terminals():
            if term.branch_length != None:
                term.branch_length = term.branch_length*factor
        return tree



from .base import Tree


class TotalTreeLength(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        total_tree_length = self.calculate_total_tree_length(tree)
        print(round(total_tree_length, 4))


    def process_args(self, args):
        return dict(tree_file_path=args.tree)

    def calculate_total_tree_length(self, tree):
        total_len = float(0.0)
        total_len = tree.total_branch_length()

        if isinstance(total_len, (int, float)):
            return total_len



import sys

from Bio.Align import MultipleSeqAlignment

from .base import Alignment

class ParsimonyInformative(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        alignment, _ = self.get_alignment_and_format()
        pi_sites, aln_len, pi_sites_per = self.calculate_parsimony_informative_sites(alignment)
        if (pi_sites, aln_len, pi_sites_per):
            print(f"{pi_sites}\t{aln_len}\t{round(pi_sites_per, 4)}")


    def process_args(self, args):
        return dict(alignment_file_path=args.alignment)

    def get_number_of_occurences_per_character(
        self,
        alignment,
        idx: int
    ) -> dict:

        # obtain sequence at position, remove gaps, and make
        # all characters uppercase
        seq_at_position = ""
        seq_at_position += alignment[:, idx]
        seq_at_position = seq_at_position.upper().replace("-", "")
        num_occurences = {}
        for char in set(seq_at_position.replace("-", "")):
            num_occurences[char] = seq_at_position.count(char)

        return num_occurences

    def count_if_parsimony_informative(
        self,
        num_occurences: dict,
        pi_sites: int
    ):
        # create a dictionary of characters that occur at least twice
        d = dict((k, v) for k, v in num_occurences.items() if v >= 2)

        # determine number of characters that occur at least twice
        if len(d) >= 2:
            pi_sites += 1 
        
        return pi_sites

    def calculate_parsimony_informative_sites(
        self,
        alignment
    ):
        # get aln length
        aln_len = alignment.get_alignment_length()
        
        pi_sites = 0

        # count number of parsimony informative sites
        for idx in range(0, aln_len, int(1)):
            # count occurneces of each character at site idx
            num_occurences = self.get_number_of_occurences_per_character(alignment, idx) 

            # add one to pi_sites if site idx is parsimony informative
            pi_sites = self.count_if_parsimony_informative(num_occurences, pi_sites)

        # calculate percent of variable sites
        pi_sites_per = (pi_sites / aln_len)*100
        
        return pi_sites, aln_len, pi_sites_per


from .alignment_length import AlignmentLength
from .alignment_length_no_gaps import AlignmentLengthNoGaps
from .column_score import ColumnScore
from .create_concatenation_matrix import CreateConcatenationMatrix
from .dna_threader import DNAThreader
from .faidx import Faidx
from .gc_content import GCContent
from .pairwise_identity import PairwiseIdentity
from .parsimony_informative_sites import ParsimonyInformative
from .rcv import RelativeCompositionVariability
from .rename_fasta_entries import RenameFastaEntries
from .sum_of_pairs_score import SumOfPairsScore
from .variable_sites import VariableSites



import sys

from ..base import BaseService
from ...helpers.files import get_alignment_and_format as get_alignment_and_format_helper

class Alignment(BaseService):
    def __init__(
        self,
        *args,
        alignment_file_path=None,
        fasta=None,
        output_file_path=None,
        protein_file_path=None,
        nucleotide_file_path=None,
        alignment_list_path=None,
        prefix=None,
        idmap=None,
        reference=None,
        verbose=None,
        entry=None
    ):
        self.alignment_file_path = alignment_file_path
        self.output_file_path = output_file_path
        self.protein_file_path = protein_file_path,
        self.nucleotide_file_path = nucleotide_file_path 
        self.alignment_list_path = alignment_list_path
        self.prefix = prefix
        self.fasta = fasta
        self.idmap = idmap
        self.reference = reference
        self.verbose = verbose
        self.entry = entry

    def get_alignment_and_format(self):
        """
        automatic file type determination
        """
        try:
            return get_alignment_and_format_helper(self.alignment_file_path)
        except FileNotFoundError:
            print("Input corresponds to no such file or directory.")
            print("Please double check pathing and filenames")
            sys.exit()
            
    def calculate_rcv(self):
        alignment, _ = self.get_alignment_and_format()
        aln_len = alignment.get_alignment_length()

        # string to hold all sequences
        concat_seq = ''
        # initialize a counter for the number of sequences in the input fasta file
        num_records = 0

        # for each record join concatSeq string and sequence as well as keeping track 
        # of the number of records
        for record in alignment:
            concat_seq  += record.seq
            num_records += 1

        # dictionary to hold the average occurence of each sequence letter
        average_d = {}
        # loop through the different sequences that appear in the fasta file
        # population dictionary with how many times that sequence appears
        for seq in set(concat_seq):
            average_d[seq] = (concat_seq.count(seq)/num_records)

        # intiailize list to hold the RCV values per ith taxa 
        # that will later be summed
        indiv_rcv_values = []

        # loop through records again and calculate RCV for 
        # each taxa and append to indivRCVvalues
        for record in alignment:
            # temp holds a temporary value of the numerator before appending
            # to numeratorRCVvalues and then is reassigned to 0 when it goes
            # through the loop again
            temp = 0
            # calculates the absolute value of the ith sequence letter minus the average
            for seq_letter in set(concat_seq):
                temp += abs(record.seq.count(seq_letter)-average_d[seq_letter])
            indiv_rcv_values.append(temp/(num_records*aln_len))

        relative_composition_variability = sum(indiv_rcv_values)

        # print the sum of all RCV values
        return relative_composition_variability




import sys
import getopt
import os.path
import statistics as stat
from typing import Tuple

from Bio import Phylo
from Bio.Phylo.BaseTree import TreeMixin
import itertools
import numpy as np

from .base import Tree

from ...helpers.stats_summary import calculate_summary_statistics_from_arr, print_summary_statistics

class PatristicDistances(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        patristic_distances, combos, stats = self.calculate_patristic_distances(tree)
        if self.verbose:
            try:
                for combo, patristic_distance in zip(combos, patristic_distances):
                    print(f"{combo[0]}-{combo[1]}\t{round(patristic_distance, 4)}")
            except BrokenPipeError:
                pass
        else:
            print_summary_statistics(stats)

    def process_args(self, args):
        return dict(tree_file_path=args.tree, verbose=args.verbose)

    def calculate_distance_between_pairs(self, tips: list, tree) -> Tuple[list, list]:
        # determine pairwise combinations of tips
        combos = list(itertools.combinations(tips, 2))

        # determine average distance between tips
        patristic_distances = []
        for combo in combos:
            patristic_distances.append(tree.distance(combo[0], combo[1]))

        return combos, patristic_distances

    def calculate_patristic_distances(self, tree):
        # get tree tips
        tips = self.get_tip_names_from_tree(tree)
        
        # get distances between pairs of taxa
        combos, patristic_distances = self.calculate_distance_between_pairs(tips, tree)
        
        # calculate summary stats
        stats = calculate_summary_statistics_from_arr(patristic_distances)
        
        return patristic_distances, combos, stats


from .base import Tree


class Treeness(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        treeness = self.calculate_treeness(tree)
        print(round(treeness, 4))


    def process_args(self, args):
        return dict(tree_file_path=args.tree)



import statistics as stat
import sys
from textwrap import dedent
from typing import Tuple

from Bio import SeqIO, SeqRecord
from Bio.Seq import Seq

from .base import Alignment

from ...helpers.files import read_single_column_file_to_list

class CreateConcatenationMatrix(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        self.create_concatenation_matrix(self.alignment_list_path, self.prefix)

    def process_args(self, args) -> dict:
        return dict(
            alignment_list_path=args.alignment_list,
            prefix=args.prefix
        )

    def read_alignment_paths(
        self,
        alignment_list_path: str
    ) -> list:
        """
        Read alignment paths into a list 
        """
        try:
            return read_single_column_file_to_list(alignment_list_path)
        except FileNotFoundError:
            print("Alignment list file (-a) is not found. Please check pathing.")
            sys.exit()

    def get_taxa_names(
        self,
        alignment_paths: list
    ) -> list:
        # get taxa names
        taxa = set()
        for alignment_path in alignment_paths:
            alignment = SeqIO.parse(alignment_path, "fasta")
            for seq_record in alignment:
                taxa.add(seq_record.name)

        return list(sorted(taxa))

    def print_start_message(
        self,
        taxa: list,
        alignment_paths: list,
        file_partition: str,
        fasta_output: str,
        file_occupancy: str
    ) -> None:
        """
        print a log message to the user about general features of
        input and output file names
        """

        start_message = dedent(f"""
            --------------------
            | General features |
            --------------------
            Total number of taxa: {len(taxa)}
            Total number of alignments: {len(alignment_paths)}


            ----------------
            | Output files |
            ----------------
            Partition file output: {file_partition}
            Concatenated fasta output: {fasta_output}
            Occupancy report: {file_occupancy}
        """)

        print(start_message)



    def get_list_of_taxa_and_records(
        self,
        alignment_path: str
    ) -> Tuple[list, list]:
        """
        get list of a taxa and their records
        """
        # list for keeping track of taxa per USCO
        og_taxa = []
        # open alignment file
        records = list(SeqIO.parse(alignment_path, "fasta"))

        # create lists of USCOtaxa and missing_taxa
        for record in records:
            # append taxa per USCO to USCOtaxa
            og_taxa.append(record.id)

        return og_taxa, records

    def determine_missing_sequences(
        self,
        taxa: list,
        og_taxa: list
    ) -> list:
        """
        determine taxa not present in an alignment file
        """
        missing_taxa = list(set(taxa)-set(og_taxa))

        return missing_taxa

    def create_missing_seq_str(
        self,
        records: list
    ) -> Tuple[str, int]:
        """
        create a placeholder string for sequences with missing taxa
        """
        og_len = ''
        missing_seq = ''
        og_len = len(records[0].seq)
        missing_seq = og_len * '?'
    
        return missing_seq, og_len

    def create_missing_taxa_sequence(
        self,
        missing_taxa: list,
        missing_seq: str,
        concat: dict
    ) -> dict:

        for indiv in missing_taxa:
            concat[indiv].append(missing_seq)
            
        return concat 

    def create_record_for_present_taxa(
        self,
        records: list,
        og_taxa: list,
        concat: dict
    ) -> dict:
        """
        create a record for taxa that are present in the concatenation dict
        """
        for record in records:
            if record.id in og_taxa:
                concat[record.id].append(record.seq)

        return concat

    def add_to_partition_info(
        self,
        partition_info: list,
        og_len: int,
        field_one: str,
        fasta: str,
        first_len: int,
        second_len: int
    ) -> Tuple[list, int, int]:
        """
        append to partition file
        """
        # second value in partition file
        second_len += og_len
        # add partition file entry to partition_info
        entry = f"{field_one}, {str(fasta)}={str(first_len)}-{str(second_len)}\n"
        partition_info.append(entry)
        # add to first value for partition file
        first_len += og_len

        return partition_info, first_len, second_len

    def add_to_occupancy_info(
        self,
        occupancy_info: list,
        og_taxa: list,
        missing_taxa: list,
        fasta: str
    ) -> list:
        """
        write partition file to detail boundaries in the concat alignment
        """

        # determine number of present taxa
        num_present = len(og_taxa)
        # determine number of missing taxa
        num_missing = len(missing_taxa)
        # determine percent occupied
        percent_occupancy = num_present/(num_present+num_missing)
        # handle missing taxa
        if num_missing == 0:
            missing_taxa = ["None"]
        entry = f"{str(fasta)}\t{str(num_present)}\t{str(num_missing)}\t{str(percent_occupancy)}\t{';'.join(missing_taxa)}\t{';'.join(og_taxa)}\n"
        occupancy_info.append(entry)
        
        return occupancy_info

    def fasta_file_write(
        self,
        fasta_output: str,
        concat: dict
    ) -> None:
        """
        join seqs of genes in value (list) and write to concat_fa 
        """
        with open(fasta_output, "w") as final_fasta_file: 
            for x in concat:
                concatenated = []
                for s in concat[x]:
                    try:
                        # if a seq object
                        concatenated.append(s._data)
                    except AttributeError:
                        # if a string
                        concatenated.append(s)
                concat[x] = concatenated
                entry = f">{x}\n{''.join(concat[x])}\n"
                final_fasta_file.write(str(entry))   

    def write_occupancy_or_partition_file(
        self,
        list_of_lists: list,
        output_file_name: str
    ) -> None:
        """
        loops through occupancy or partition file information and writes
        them to the corresponding output file
        """

        with open(output_file_name, "w") as f:
            for entry in list_of_lists:
                f.write(entry)

    def create_concatenation_matrix(
        self,
        alignment_list_path: str,
        prefix: str
    ) -> None:
        alignment_paths = self.read_alignment_paths(alignment_list_path)

        taxa = self.get_taxa_names(alignment_paths)

        # assign the output file names
        file_partition = f"{prefix}.partition"
        fasta_output = f"{prefix}.fa"
        file_occupancy = f"{prefix}.occupancy"

        # print log message
        self.print_start_message(
            taxa,
            alignment_paths,
            file_partition,
            fasta_output,
            file_occupancy
        )

        # assign placeholders for lengths in the partition file
        first_len    = 1
        second_len   = 0

        # create a dictionary with keys as taxa and empty list values
        concatenated_seqs = {key: [] for key in taxa}

        # string for first field of partition file
        field_one = 'AUTO'

        # create empty lists to contain information for the 
        # partition and occupancy files
        partition_info = []
        occupancy_info = []

        # loop through alignment files
        for alignment_path in alignment_paths:

            # create a list of taxa and records
            og_taxa, records = self.get_list_of_taxa_and_records(
                alignment_path
            )
                
            # initialize list and determine missing taxa
            missing_taxa = self.determine_missing_sequences(
                taxa,
                og_taxa
            )

            # create string for missing seq
            missing_seq, og_len = self.create_missing_seq_str(
                records
            )

            # create missing taxa sequence
            concatenated_seqs = self.create_missing_taxa_sequence(
                missing_taxa,
                missing_seq,
                concatenated_seqs
            )      

            # create record for present data
            concatenated_seqs = self.create_record_for_present_taxa(
                records,
                og_taxa,
                concatenated_seqs
            )

            # append partition to to partition info
            partition_info, first_len, second_len = self.add_to_partition_info(
                partition_info,
                og_len,
                field_one,
                alignment_path,
                first_len,
                second_len
            )

            # append occupancy data to occupancy info
            occupancy_info = self.add_to_occupancy_info(
                occupancy_info,
                og_taxa,
                missing_taxa,
                alignment_path
            )

        # write output fasta file
        self.fasta_file_write(
            fasta_output,
            concatenated_seqs
        )

        # write out occupancy and partition file
        self.write_occupancy_or_partition_file(occupancy_info, file_occupancy)
        self.write_occupancy_or_partition_file(partition_info, file_partition)

        print("Complete!\n")



import sys
from typing import Tuple

from Bio import Phylo

from ..base import BaseService

class Tree(BaseService):
    def __init__(
        self,
        *args,
        tree_file_path=None,
        idmap=None,
        alignment_file_path=None,
        tree1_file_path=None,
        outgroup_taxa_file_path=None,
        output_file_path=None,
        factor=None,
        remove=None,
        verbose=None,
        reference=None,
        list_of_taxa=None,
        trees=None,
        groups=None,
        support=None
    ):
        self.tree_file_path = tree_file_path
        self.tree1_file_path = tree1_file_path
        self.alignment_file_path = alignment_file_path
        self.output_file_path = output_file_path
        self.outgroup_taxa_file_path = outgroup_taxa_file_path
        self.tree_format = "newick"
        self.verbose = verbose
        self.factor = factor
        self.remove = remove
        self.idmap = idmap
        self.reference = reference
        self.list_of_taxa = list_of_taxa
        self.trees = trees
        self.groups = groups
        self.support = support


    def read_tree_file(self):
        try:
            return Phylo.read(self.tree_file_path, self.tree_format)
        except FileNotFoundError:
            print(f"{self.tree_file_path} corresponds to no such file or directory.")
            print("Please check filename and pathing")
            sys.exit()


    def read_tree1_file(self):
        try:
            return Phylo.read(self.tree1_file_path, self.tree_format)
        except FileNotFoundError:
            print(f"{self.tree1_file_path} corresponds to no such file or directory.")
            print("Please check filename and pathing")
            sys.exit()


    def read_reference_tree_file(self):
        try:
            return Phylo.read(self.reference, self.tree_format)
        except FileNotFoundError:
            print(f"{self.reference} corresponds to no such file or directory.")
            print("Please check filename and pathing")
            sys.exit()


    def write_tree_file(self, tree, output_file_path):
        return Phylo.write(tree, output_file_path, self.tree_format)

    def get_tip_names_from_tree(
        self, tree
        ) -> list:
        """
        get tip names from a tree
        """
        tips = []
        for tip in tree.get_terminals():
            tips.append(tip.name)
        
        return tips

    def shared_tips(
        self, 
        a,
        b
        ):
        """
        Determines what tips are shared between two trees
        -------------------------------------------------
        argv: a
            list of tips from one tree
        argv: b
            list of tips from a second tree
        """ 

        a_set = set(a) 
        b_set = set(b) 
        
        # check length  
        if len(a_set.intersection(b_set)) > 0: 
            return(list(a_set.intersection(b_set)))   
        else: 
            print("no common tips") 
            sys.exit()

    def prune_tree_using_taxa_list(
        self,
        tree,
        taxa_to_prune: list
    ):
        """
        prune taxa from tree
        """
        for taxon in taxa_to_prune:
            tree.prune(taxon)
        
        return tree

    def calculate_treeness(self, tree=None, print_value=False):
        if not tree:
            tree = self.read_tree_file()

        inter_len = float(0.0)
        # determine internal branch lengths
        for interal in tree.get_nonterminals():
            # only include if a branch length value is present
            if interal.branch_length != None:
                inter_len += interal.branch_length
        # determine total branch length
        total_len = tree.total_branch_length()

        try:
            treeness = float(inter_len / total_len)
            try:
                if print_value:
                    print(f"{treeness}")
                return treeness
            except BrokenPipeError:
                pass
        except ZeroDivisionError:
            try:
                print("Invalid tree. Tree should contain branch lengths")
                return None
            except BrokenPipeError:
                pass



import getopt
import logging
import os.path
import sys

from Bio import Phylo

from .base import Tree

class CollapseBranches(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        tree.collapse_all(lambda c: c.confidence is not None and c.confidence < self.support)
        self.write_tree_file(tree, self.output_file_path)

    def process_args(self, args):
        if args.output is None:
            output_file_path = f"{args.tree}.collapsed_{args.support}.tre"
        else:
            output_file_path = f"{args.output}"
        return dict(
            tree_file_path=args.tree,
            support=args.support,
            output_file_path=output_file_path,
        )


import sys
import getopt
import os.path
import statistics as stat

from Bio import Phylo
from Bio.Phylo.BaseTree import TreeMixin
import itertools
import numpy as np

from .base import Tree

from ...helpers.stats_summary import calculate_summary_statistics_from_arr, print_summary_statistics


class LBScore(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        tips, LBis = self.calculate_lb_score(tree)
        if self.verbose:
            try:
                for tip, LBi in zip(tips, LBis):
                    print(f"{tip}\t{round(LBi, 4)}")
            except BrokenPipeError:
                pass
        else:
            stats = calculate_summary_statistics_from_arr(LBis)
            print_summary_statistics(stats)

    def process_args(self, args):
        return dict(tree_file_path=args.tree, verbose=args.verbose)

    def calculate_average_distance_between_tips(
        self,
        tips: list,
        tree
    ) -> float:
        # determine pairwise combinations of tips
        combos = list(itertools.combinations(tips, 2))

        # determine average distance between tips
        # avg_dist is PDa
        total_dist = float()
        for combo in combos:
            total_dist+=tree.distance(combo[0], combo[1])
        
        return total_dist/len(combos)

    def calculate_average_distance_of_taxon_to_other_taxa(
        self,
        tips: list,
        tree
    ) -> list:
        """
        calculate average distance of taxon to all other taxon or average PDi.
        Save results to avg PDis list
        """
        avg_PDis = []
        for tip in tips:
            tips_minus_i = list(set(tips) - set(tip))
            PDi = []
            for tip_minus in tips_minus_i:    
                PDi.append(tree.distance(tip, tip_minus))
            PDi = sum(PDi) / len(PDi)
            avg_PDis.append(PDi)

        return avg_PDis

    def calculate_lb_score_per_taxa(
        self,
        avg_PDis:list,
        avg_dist:float
    ) -> list:
        """
        create a list with the lb scores for each taxon
        """
        LBis = []
        for PDi in avg_PDis:
            try:
                LBis.append((((PDi/avg_dist)-1)*100))
            except ZeroDivisionError:
                try:
                    print("Invalid tree. Tree should contain branch lengths")
                    sys.exit()
                except BrokenPipeError:
                    pass
        
        return LBis

    def calculate_lb_score(self, tree):
        # get tree tips
        tips = self.get_tip_names_from_tree(tree)
        
        # get average distance between tips
        avg_dist = self.calculate_average_distance_between_tips(tips, tree)

        # calculate average distance of taxon i to all other taxa
        # or PDi and save each result to LBi
        avg_PDis = self.calculate_average_distance_of_taxon_to_other_taxa(tips, tree)

        # use PDis and avgDist to calculate LB values for each taxon
        LBis = self.calculate_lb_score_per_taxa(avg_PDis, avg_dist)
        
        return tips, LBis



import itertools
import statistics as stat

from Bio.Align import MultipleSeqAlignment
import numpy as np

from .base import Alignment
from ...helpers.stats_summary import calculate_summary_statistics_from_dict, print_summary_statistics

class PairwiseIdentity(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # get aln
        alignment, alignment_format = self.get_alignment_and_format()
        
        # get entry indices
        entries = self.get_entry_indices(alignment)

        # determine pairwise combinations of entry indices
        combos = list(itertools.combinations(entries, 2))

        pairwise_identities, stats = self.calculate_pairwise_identities(alignment, combos)
        
        if self.verbose:
            try:
                for pair, identity in pairwise_identities.items():
                    print(f"{pair}\t{identity}")
            except BrokenPipeError:
                pass
        else:
            print_summary_statistics(stats)

    def process_args(self, args):
        return dict(alignment_file_path=args.alignment, verbose=args.verbose)

    def get_entry_indices(self, alignment) -> list:
        entries = []
        entries_count = 0
        for record in alignment:
            entries.append(entries_count)
            entries_count+=1

        return entries

    def calculate_pairwise_identities(self, alignment, combos):
        # get aln length
        aln_len = alignment.get_alignment_length()

        # determine pairwise identity of each combination
        pairwise_identities = {}
        for combo in combos:
            identities = 0
            seq_one = alignment[combo[0]].seq
            seq_two = alignment[combo[1]].seq
            for idx in range(0, aln_len):
                if seq_one[idx] == seq_two[idx]:
                    identities += 1
            ids = alignment[combo[0]].id + '-' + alignment[combo[1]].id
            pairwise_identities[ids] = identities / aln_len

        stats = calculate_summary_statistics_from_dict(pairwise_identities)

        return pairwise_identities, stats


import getopt
import logging
import os.path
import sys

from scipy.stats import zscore
from scipy.stats.stats import pearsonr

from Bio import Phylo
from Bio.Phylo.BaseTree import TreeMixin

from .base import Tree

class CovaryingEvolutionaryRates(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree_zero = self.read_tree_file()
        tree_one = self.read_tree1_file()
        tree_ref = self.read_reference_tree_file()

        ## - Calculate correlation between two gene trees 
        ## and save results to an array, corrArr.
        ## - Branch lengths will also be part of output

        # get tree tip names
        tree_zero_tips = self.get_tip_names_from_tree(tree_zero)
        tree_one_tips = self.get_tip_names_from_tree(tree_one)
        tree_ref_tips = self.get_tip_names_from_tree(tree_ref)

        # get shared tips between the two trees
        shared_tree_tips = self.shared_tips(tree_zero_tips, tree_one_tips)

        # find differences between tree tips and shared tips
        # to determine what tips to prune
        tree_zero_tips_to_prune   = list(set(tree_zero_tips) - set(shared_tree_tips))
        tree_one_tips_to_prune    = list(set(tree_one_tips) - set(shared_tree_tips))
        tree_ref_tips_to_prune    = list(set(tree_ref_tips) - set(shared_tree_tips))

        # get a set of pruned trees
        tree_zero = self.prune_tips(tree_zero, tree_zero_tips_to_prune)
        tree_one = self.prune_tips(tree_one, tree_one_tips_to_prune)
        tree_ref = self.prune_tips(tree_ref, tree_ref_tips_to_prune)

        # check that the input trees have the same topology
        self.determine_if_trees_differ(tree_zero, tree_one, tree_ref)

        # obtain corrected branch lengths where branch lengths
        # are corrected by the species tree branch length
        tree_zero_corr_branch_lengths = self.correct_branch_lengths(tree_zero, tree_ref)
        tree_one_corr_branch_lengths = self.correct_branch_lengths(tree_one, tree_ref)

        # remove corrected BLs greater than 5
        outlier_indices = []
        outlier_indices = self.get_indices_of_outlier_branch_lengths(tree_zero_corr_branch_lengths, outlier_indices)
        outlier_indices = self.get_indices_of_outlier_branch_lengths(tree_one_corr_branch_lengths, outlier_indices)

        tree_zero_corr_branch_lengths = self.remove_outliers_based_on_indices(tree_zero_corr_branch_lengths, outlier_indices)
        tree_one_corr_branch_lengths = self.remove_outliers_based_on_indices(tree_one_corr_branch_lengths, outlier_indices)

        # standardize values for final correction
        tree_zero_corr_branch_lengths = zscore(tree_zero_corr_branch_lengths)
        tree_one_corr_branch_lengths = zscore(tree_one_corr_branch_lengths)

        # Calculate correlation and append to results array
        # also keep a list of p values
        corr = (list(pearsonr(tree_zero_corr_branch_lengths, tree_one_corr_branch_lengths)))
        
        try:
            if self.verbose:
                for val_zero, val_one in zip(tree_zero_corr_branch_lengths, tree_one_corr_branch_lengths):
                    print(f"{round(val_zero, 4)}\t{round(val_one, 4)}")
            else:
                print(f"{round(corr[0], 4)}\t{round(corr[1], 6)}")
        except BrokenPipeError:
            pass

    def process_args(self, args):
        return dict(
            tree_file_path=args.tree_zero,
            tree1_file_path=args.tree_one,
            reference=args.reference,
            verbose=args.verbose
        )

    def get_indices_of_outlier_branch_lengths(self, corr_branch_lengths, outlier_indices):
        """
        create index for branch lengths that 
        have an absolute value greater than 5
        """
        for idx in range(0, len(corr_branch_lengths)): 
            if corr_branch_lengths[idx] > 5 or corr_branch_lengths[idx] < -5:
                if idx not in outlier_indices: 
                    outlier_indices.append(idx)
        
        return outlier_indices

    def remove_outliers_based_on_indices(self, corr_branch_lengths, outlier_indices):
        """
        remove value if the value is an outlier according
        to the outlier indices list
        """
        corr_branch_lengths = [i for j, i in enumerate(corr_branch_lengths) if j not in outlier_indices]
        return corr_branch_lengths

    # def shared_tips(
    #     self, 
    #     a,
    #     b
    #     ):
    #     """
    #     Determines what tips are shared between two trees
    #     -------------------------------------------------
    #     argv: a
    #         list of tips from one tree
    #     argv: b
    #         list of tips from a second tree
    #     """ 

    #     a_set = set(a) 
    #     b_set = set(b) 
        
    #     # check length  
    #     if len(a_set.intersection(b_set)) > 0: 
    #         return(list(a_set.intersection(b_set)))   
    #     else: 
    #         print("no common tips") 
    #         sys.exit()
 
    
    def prune_tips(
        self, tree, tips
        ):
        """
        prune tips from trees
        """

        for tip in tips:
            tree.prune(tip)
        
        return tree

    def determine_if_trees_differ(self, tree_zero, tree_one, tree_ref):
        """
        determine if the trees differ from one another
        """
        differences = 0
        differences = self.compare_trees(differences, tree_zero, tree_one)
        differences = self.compare_trees(differences, tree_zero, tree_ref)
        differences = self.compare_trees(differences, tree_one, tree_ref)
        
        if differences > 0:
            print("Input trees differ in topology.")
            print("Please ensure input phylogenies all have the same topology.")
            sys.exit()

    def compare_trees(
        self,
        plain_rf: int,
        tree_zero: Tree,
        tree_one: Tree
    ) -> int:
        # loop through tree_zero and find similar clade in tree_one
        for clade_zero in tree_zero.get_nonterminals():
            # initialize and populate a list of tip names in tree_zero
            tip_names_zero = self.get_tip_names_from_tree(clade_zero)
            # get common ancestor of tree_zero tip names in tree_one
            clade_one = tree_one.common_ancestor(tip_names_zero)
            # initialize and populate a list of tip names in tree_one
            tip_names_one = self.get_tip_names_from_tree(clade_one)

            # compare the list of tip names
            plain_rf = self.determine_if_clade_differs(
                plain_rf,
                tip_names_zero,
                tip_names_one
            )

        return plain_rf
            
    def determine_if_clade_differs(
        self,
        plain_rf: int,
        tip_names_zero: list,
        tip_names_one: list
        ) -> int:
        """
        if clade differs, add 1 to plain_rf value
        """

        if set(tip_names_zero) != set(tip_names_one):
            plain_rf +=1
        
        return plain_rf

    def correct_branch_lengths(
        self, 
        t,
        sp
        ):
        """
        obtain a list of corrected branch lengths
        """
        
        l = []
        for i, s in zip(t.get_terminals(), sp.get_terminals()):
            l.append(i.branch_length/s.branch_length)
        return(l)


import getopt
import logging
import math
import os.path
import statistics as stat
import sys
from typing import Tuple

from Bio import Phylo
from Bio.Phylo.BaseTree import TreeMixin
import itertools
import numpy as np

from .base import Tree

from ...helpers.files import read_single_column_file_to_list

class DVMC(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        outgroup = read_single_column_file_to_list(self.outgroup_taxa_file_path)
        dvmc = self.determine_dvmc(tree, outgroup)
        
        print(round(dvmc, 4))


    def process_args(self, args):
        return dict(tree_file_path=args.tree, outgroup_taxa_file_path=args.root)

    def get_names_of_outgroup_taxa_that_are_present(
        self,
        outgroup: list,
        tree: Tree
    ) -> list:
        # initialize list for outgroup taxa that are present in tree
        out_pres = []
        # loop through terminal branch
        for term in tree.get_terminals():
            # if outgroup taxa is present, append it to out_pres
            if term.name in outgroup:
                out_pres.append(term.name)

        return out_pres

    def get_term_to_root_dist_and_sum_of_distances(
        self,
        tree
    ) -> Tuple[float, float]:
        """
        calculate root to tip distances and
        the sum of root to tip distances
        """ 
        dist = []
        sum_dist = 0
        for term in tree.get_terminals():
            # append root to tip distance to dist
            dist.append((tree.distance(term)))
            # keep running sum of tree distances
            sum_dist += tree.distance(term)

        return dist, sum_dist

    def calculate_dvmc(
        self,
        dist: float,
        sum_dist: float,
        num_spp: int
    ) -> float:
        """
        calculate dvmc from tip to root distances
        """
        # determine dvmc
        # calculate average tree distance
        avg_dist = sum_dist/num_spp

        # determine the sum of i=1 to N for (x_i-x_bar)^2
        sumi2N = float(0.0)
        for x_i in dist:
            sumi2N += ((x_i-avg_dist)**2)

        # multiply sumi2N by 1/(N-1) where N is the number of spp
        # and then take the square root
        dvmc = float(0.0)
        dvmc = math.sqrt((1/(num_spp-1))*sumi2N) 

        return dvmc

    def determine_dvmc(self, tree: Tree, outgroup: list):
        # get names of outgroup taxa in tree
        out_pres = self.get_names_of_outgroup_taxa_that_are_present(outgroup, tree)

        # root tree on outgroup
        tree.root_with_outgroup(out_pres)

        # prune outgroup taxa from tree
        tree = self.prune_tree_using_taxa_list(tree, out_pres)

        # loop through terminal branches and store 
        # distances from the root to the tip in a list.
        # Also, calc the sum of all tip to root distances
        dist, sum_dist = self.get_term_to_root_dist_and_sum_of_distances(tree)

        # determine number of taxa in the tree
        num_spp = tree.count_terminals()

        # calculate and return dvmc
        return self.calculate_dvmc(dist, sum_dist, num_spp)
        


from enum import Enum
import re
import sys

from Bio import SeqIO

from .base import Alignment

from ...helpers.files import get_alignment_and_format

class FileFormat(Enum):
    fasta = "fasta"
    clustal = "clustal"
    maf = "maf"
    mauve = "mauve"
    phylip = "phylip"
    phylip_seq = "phylip-sequential"
    stockholm = "stockholm"

class GCContent(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # create biopython object of sequences
        records, _ = get_alignment_and_format(self.fasta)
        
        # initialize and populate dict for
        # holding the entry sequences
        entry_and_seq = {}
        for record in records:
            entry_and_seq[record.id] = str(record.seq)

        if self.verbose:
            for entry, seq in entry_and_seq.items():
                seq, matches = self.find_matches_and_remove_gaps(seq)
                try:
                    print(f"{entry}\t{round(len(matches)/len(seq), 4)}")
                except BrokenPipeError:
                    pass
        else:
            all_seqs = []
            for entry, seq in entry_and_seq.items():
                all_seqs.append(seq)
            all_seqs = ''.join(all_seqs)
            all_seqs, matches = self.find_matches_and_remove_gaps(all_seqs)
            try:
                gc_content = round(len(matches)/len(all_seqs), 4)
            except ZeroDivisionError:
                try:
                    print("Input file has an unacceptable format. Please check input file argument.")
                    sys.exit()
                except BrokenPipeError:
                    pass
            try:
                print(gc_content)
            except BrokenPipeError:
                pass
            

    def process_args(self, args):
        return dict(
            fasta=args.fasta, 
            verbose=args.verbose
        )

    def find_matches_and_remove_gaps(self, seq: str):
        regex_pattern = re.compile('[GgCc]')
        seq = seq.replace('-', '')
        seq = seq.replace('?', '')
        matches = regex_pattern.findall(seq)
        return seq, matches



import logging

from Bio import Phylo

from .base import Tree

from ...helpers.files import read_single_column_file_to_list

class PruneTree(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        
        taxa = read_single_column_file_to_list(self.list_of_taxa)

        tree = self.prune_tree_using_taxa_list(tree, taxa)

        self.write_tree_file(tree, self.output_file_path)
    
    def process_args(self, args):
        tree_file_path = args.tree
        if args.output is None:
            output_file_path = f"{tree_file_path}.pruned"
        else:
            output_file_path = f"{args.output}"

        return dict(
            tree_file_path=tree_file_path,
            list_of_taxa=args.list_of_taxa,
            output_file_path=output_file_path,
        )


import sys

from Bio import SeqIO

from .base import Alignment

class RenameFastaEntries(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # create biopython object of sequences
        records = SeqIO.parse(self.fasta, "fasta")

        # save idmap to a dictionary
        idmap = self.idmap_to_dictionary(self.idmap)

        # replace and write out
        self.replace_ids_and_write(self.output_file_path, records, idmap)

        
    def process_args(self, args):
        if args.output is None:
            output_file_path = f"{args.fasta}.renamed.fa"
        else:
            output_file_path = f"{args.output}"

        return dict(
            fasta=args.fasta, 
            idmap=args.idmap,
            output_file_path=output_file_path
        )

    def replace_ids_and_write(self, output_file_path, records, idmap):
        """
        for tips with a name as a key in the idmap,
        replace that tip name with the value in the
        idmap and write to output file
        """
        with open(output_file_path, 'w') as output_file_path:
            try:
                for record in records:
                    if record.id in idmap:
                        # replace ID
                        record.id = idmap[record.id]
                        # remove description
                        record.description = ''
                    SeqIO.write(record, output_file_path, "fasta")
            except FileNotFoundError:
                print(f"{self.fasta} corresponds to no such file or directory.")
                print("Please double check pathing and filenames")
                sys.exit()


    def idmap_to_dictionary(self, idmap:str) -> dict:
        """
        read idmap into a dictionary
        """
        idmap={}
        try:
            with open(self.idmap) as identifiers:
                for line in identifiers:
                    (key, val) = line.split()
                    idmap[key] = val
            return idmap
        except FileNotFoundError:
            print(f"{self.idmap} corresponds to no such file or directory.")
            print("Please double check pathing and filenames")
            sys.exit()




from enum import Enum
import itertools
from typing import Tuple

from Bio import AlignIO
import scipy

from .base import Tree
from ...helpers.files import get_alignment_and_format as get_alignment_and_format_helper

class FileFormat(Enum):
    fasta = "fasta"
    clustal = "clustal"
    maf = "maf"
    mauve = "mauve"
    phylip = "phylip"
    phylip_seq = "phylip-sequential"
    phylip_rel = "phylip-relaxed"
    stockholm = "stockholm"

class Saturation(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # read in alignment
        alignment, alignment_format = get_alignment_and_format_helper(self.alignment_file_path)
        
        # read in tree
        tree = self.read_tree_file()

        # get tip and tip combinations
        tips = self.get_tip_names_from_tree(tree)
        combos = list(itertools.combinations(tips, 2))

        # for pairwise combinations, calculate patristic
        # distances and pairwise identities
        patristic_distances, pairwise_identities = self.loop_through_combos_and_calculate_pds_and_pis(
            combos,
            alignment,
            tree
        )
            
        # calculate linear regression
        _, _, r_value, _, _ = scipy.stats.linregress(pairwise_identities, patristic_distances)
    
        # report res
        self.print_res(
            self.verbose,
            combos,
            pairwise_identities,
            patristic_distances,
            r_value
        )

    def process_args(self, args):
        return dict(
            tree_file_path=args.tree,
            alignment_file_path=args.alignment,
            verbose=args.verbose
        )

    def loop_through_combos_and_calculate_pds_and_pis(
        self,
        combos: list, 
        alignment,
        tree
    ) -> Tuple[list, list]:
        """
        loop through all taxon combinations and determine
        their patristic distance and pairwise identity
        """
        patristic_distances = []
        pairwise_identities = []
        aln_len = alignment.get_alignment_length()
        for combo in combos:
            # calculate pd
            patristic_distances.append(tree.distance(combo[0], combo[1]))
            # calculate pairwise identity
            pairwise_identities = self.calculate_pairwise_identities(
                alignment, pairwise_identities, aln_len, combo
            )
        return patristic_distances, pairwise_identities

    def calculate_pairwise_identities(
        self,
        alignment,
        pairwise_identities: list,
        aln_len: int,
        combo: tuple
    ) -> list:
        """
        calculate pairwise identities for a given combo
        """
        identities = 0
        seq_one = ''
        seq_two = ''
        for record in alignment:
            if record.name == combo[0]:
                seq_one = record.seq
            elif record.name == combo[1]:
                seq_two = record.seq
        for idx in range(0, aln_len):
            if seq_one[idx] == seq_two[idx]:
                identities += 1
        pairwise_identities.append(identities / aln_len)

        return pairwise_identities

    def print_res(
        self,
        verbose: bool,
        combos: list,
        pairwise_identities: list,
        patristic_distances: list,
        r_value: float
    ) -> None:
        """
        print results to stdout
        """
        try:
            if verbose:
                for combo, pairwise_identity, patristic_distance in zip(combos, pairwise_identities, patristic_distances):
                    print(f"{combo[0]}-{combo[1]}\t{round(pairwise_identity,4)}\t{round(patristic_distance, 4)}")
            else:
                print(round(r_value**2, 4))
        except BrokenPipeError:
            pass



from Bio.Align import MultipleSeqAlignment

from .base import Alignment

class AlignmentLengthNoGaps(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        alignment, alignment_format = self.get_alignment_and_format()
        aln_len_no_gaps, aln_len, aln_len_no_gaps_per = self.calculate_alignment_length_no_gaps(alignment)
        print(f"{aln_len_no_gaps}\t{aln_len}\t{round(aln_len_no_gaps_per, 4)}")

    def process_args(self, args):
        return dict(alignment_file_path=args.alignment)

    def calculate_alignment_length_no_gaps(self, alignment):
        aln_len = alignment.get_alignment_length()
        aln_len_no_gaps = self.get_sites_no_gaps_count(alignment, aln_len)

        # calculate percent of variable sites
        aln_len_no_gaps_per = (aln_len_no_gaps / aln_len) * 100
        
        return aln_len_no_gaps, aln_len, aln_len_no_gaps_per

    def get_sites_no_gaps_count(self, alignment, aln_len):
        """
        Count sites in the alignment with no gaps
        """
        aln_len_no_gaps = 0
        for i in range(0, aln_len):
            seq_at_position = ""
            seq_at_position += alignment[:, i]
            if "-" not in seq_at_position:
                aln_len_no_gaps += 1

        return aln_len_no_gaps


import sys
from typing import Tuple

from Bio import SeqIO, SeqRecord

from .base import Alignment


class DNAThreader(Alignment):
    """
    Threads DNA on top of protein alignment
    """

    def __init__(self, args) -> None:
        self.process_args(args)

    def process_args(self, args):
        self.include_stop_codon = args.stop
        self.protein_file_path = args.protein
        self.nucleotide_file_path = args.nucleotide
        
    def run(self):
        prot = self.read_file(self.protein_file_path)
        nucl = self.read_file(self.nucleotide_file_path)

        pal2nal = self.thread(prot, nucl)


        for record in pal2nal:
            sequence = ''.join(pal2nal[record])
            print(f">{record}")
            print(f"{sequence}")

    def read_file(
        self,
        file_path: str,
        file_format: str = "fasta"
    ) -> SeqRecord:
        return SeqIO.parse(file_path, file_format)

    def thread(
        self,
        protein: SeqRecord,
        nucleotide: SeqRecord
    ) -> dict:
        # protein alignment to nucleotide alignment
        pal2nal = {}

        try:
            for protein_seq_record, nucleotide_seq_record in zip(protein, nucleotide):
                # get gene id and sequences
                gene_id, p_seq, n_seq = self.get_id_and_seqs(protein_seq_record, nucleotide_seq_record)

                # initialize gap counter and gene in pal2nal dict
                gap_count = 0
                pal2nal[gene_id] = []

                # loop through the sequence
                for aa_idx in range(0, len(p_seq), 1):
                    # if AA is a gap insert a codon of gaps
                    if p_seq[aa_idx] == "-":
                        pal2nal = self.add_gap(pal2nal, gene_id)
                        gap_count += 1
                    else:
                        if self.include_stop_codon:
                            # if AA is not a gap, insert the corresponding codon
                            if p_seq[aa_idx] != "-":
                                pal2nal = self.add_codon(
                                    aa_idx,
                                    gap_count,
                                    n_seq,
                                    gene_id,
                                    pal2nal
                                )
                        else:
                            # if AA is a stop or ambiguous insert a codon of gaps
                            if p_seq[aa_idx] == "X" or p_seq[aa_idx] == "*":
                                pal2nal = self.add_gap(pal2nal, gene_id)
                            else:
                                pal2nal = self.add_codon(
                                    aa_idx,
                                    gap_count,
                                    n_seq,
                                    gene_id,
                                    pal2nal
                                )
            return pal2nal
        except FileNotFoundError:
            try:
                print("One input corresponds to no such file or directory.")
                print("Please double checking pathing and filenames")
                sys.exit()
            except BrokenPipeError:
                pass

    def get_id_and_seqs(
        self,
        protein_seq_record: SeqRecord,
        nucleotide_seq_record: SeqRecord
    ):
        """
        get gene id, protein sequence, and nucleotide sequence
        """

        # save gene id to gene_id
        gene_id = protein_seq_record.id

        # save protein sequence to p_seq
        p_seq = protein_seq_record.seq

        # save nucleotide sequence to n_seq
        n_seq = nucleotide_seq_record.seq

        return gene_id, p_seq, n_seq

    def add_gap(
        self,
        pal2nal: dict,
        gene_id: str
    ) -> dict:
        """
        add a gap to the growing sequence
        """

        pal2nal[gene_id].append("---")

        return pal2nal

    def add_codon(
        self,
        aa_idx: int,
        gap_count: int,
        n_seq: SeqRecord,
        gene_id: str,
        pal2nal: dict
    ) -> dict:
        """
        add a gap to the growing sequence
        """

        nt_window = (aa_idx - gap_count) * 3
        pal2nal[gene_id].append(n_seq[nt_window : nt_window + 3]._data)

        return pal2nal








from Bio import SeqIO

from .base import Alignment

class Faidx(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        record_dict = SeqIO.index(self.fasta, "fasta")
        print(f">{record_dict[self.entry].name}\n{record_dict[self.entry].seq}")

    def process_args(self, args):
        return dict(fasta=args.fasta, entry=args.entry)



#!/usr/bin/env python

import logging
import sys
import textwrap
from .version import __version__

from argparse import (
    ArgumentParser,
    RawTextHelpFormatter,
    SUPPRESS,
    RawDescriptionHelpFormatter,
)

from .services.alignment import (
    AlignmentLength,
    AlignmentLengthNoGaps,
    ColumnScore,
    CreateConcatenationMatrix,
    DNAThreader,
    Faidx,
    GCContent,
    PairwiseIdentity,
    ParsimonyInformative,
    RelativeCompositionVariability,
    RenameFastaEntries,
    SumOfPairsScore,
    VariableSites
)

from .services.tree import (
    BipartitionSupportStats,
    BranchLengthMultiplier,
    CollapseBranches,
    CovaryingEvolutionaryRates,
    DVMC,
    EvolutionaryRate,
    InternalBranchStats,
    InternodeLabeler,
    LastCommonAncestorSubtree,
    LBScore,
    TotalTreeLength,
    PatristicDistances,
    PolytomyTest,
    PrintTree,
    PruneTree,
    RenameTreeTips,
    RobinsonFouldsDistance,
    Saturation,
    SpuriousSequence,
    TipLabels,
    Treeness,
    TreenessOverRCV
)

from .helpers.boolean_argument_parsing import str2bool


logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)

help_header = f"""
                 _____  _           _  _______ _______ 
                |  __ \| |         | |/ /_   _|__   __|
                | |__) | |__  _   _| ' /  | |    | |   
                |  ___/| '_ \| | | |  <   | |    | |   
                | |    | | | | |_| | . \ _| |_   | |   
                |_|    |_| |_|\__, |_|\_\_____|  |_|   
                               __/ |                   
                              |___/   
                            
                Version: {__version__}
                Citation: Steenwyk et al. 2021, Bioinformatics. doi: 10.1093/bioinformatics/btab096
                https://academic.oup.com/bioinformatics/advance-article-abstract/doi/10.1093/bioinformatics/btab096/6131675

"""

class Phykit(object):
    help_header = f"""
                 _____  _           _  _______ _______ 
                |  __ \| |         | |/ /_   _|__   __|
                | |__) | |__  _   _| ' /  | |    | |   
                |  ___/| '_ \| | | |  <   | |    | |   
                | |    | | | | |_| | . \ _| |_   | |   
                |_|    |_| |_|\__, |_|\_\_____|  |_|   
                               __/ |                   
                              |___/   
                            
                Version: {__version__}
                Citation: Steenwyk et al. 2021, Bioinformatics. doi: 10.1093/bioinformatics/btab096
                https://academic.oup.com/bioinformatics/advance-article-abstract/doi/10.1093/bioinformatics/btab096/6131675

    """
    
    def __init__(self):
        parser = ArgumentParser(
            add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {self.help_header}

                PhyKIT helps process and analyze multiple sequence alignments and phylogenies.

                Generally, all functions are designed to help understand the contents of alignments
                (e.g., gc content or the number of parsimony informative sites) and the shape
                of trees (e.g., treeness, degree of violation of a molecular clock).

                Some help messages indicate that summary statistics are reported (e.g., 
                bipartition_support_stats). Summary statistics include mean, median, 25th percentile,
                75th percentile, minimum, maximum, standard deviation, and variance. These functions
                typically have a verbose option that allows users to get the underlying data
                used to calculate summary statistics. 

                Usage: phykit <command> [optional command arguments]

                Command specific help messages can be viewed by adding a 
                -h/--help argument after the command. For example, to see the
                to see the help message for the command 'treeness', execute
                "phykit treeness -h" or "phykit treeness --help".

                Lastly, each function comes with aliases to save the user some
                key strokes. For example, to get the help message for the 'treeness'
                function, you can type "phykit tness -h". All aliases are specified
                in parentheses after the long form of the function name. 

                Alignment-based commands
                ========================
                alignment_length (alias: aln_len; al)
                    - calculates alignment length
                alignment_length_no_gaps (alias: aln_len_no_gaps; alng)
                    - calculates alignment length after removing sites with gaps
                column_score (alias: cs)
                    - calculate column score between a reference and query alignment
                create_concatenation_matrix (alias: create_concat; cc)
                    - create concatenation matrix from a set of alignments
                faidx (alias: get_entry; ge)
                    - extract query fasta entry from multi-fasta file              
                gc_content (alias: gc)
                    - calculate GC content of a fasta entries or entries thereof
                pairwise_identity (alias: pairwise_id, pi)
                    - calculates average pairwise identify among sequences in
                      an alignment file. This is a proxy for evolutionary rate
                parsimony_informative_sites (alias: pis)
                    - calculates the number and percentage of parsimony
                      informative sites in an alignment
                relative_composition_variability (alias: rel_comp_var, rcv)
                    - calculates relative composition variability in an alignment
                rename_fasta_entries (alias: rename_fasta)
                    - rename entries in a fasta file
                sum_of_pairs_score (alias: sops; sop)
                    - calculate sum-of-pairs score between a reference and query alignment
                thread_dna (alias: pal2nal; p2n)
                    - thread dna sequences over a protein alignment
                variable_sites (alias: vs)
                    - calculates the number and percentage of variable sites
                      in an alignment

                Tree-based commands
                ===================
                bipartition_support_stats (alias: bss)
                    - calculates summary statistics for bipartition support
                branch_length_multiplier (alias: blm)
                    - multiply all branch lengths by a specified factor
                collapse_branches (alias: collapse; cb)
                    - collapses branches according to bipartition support
                covarying_evolutionary_rates (alias: cover)
                    - calculates correlation in the evolutionary rate of two trees
                degree_of_violation_of_a_molecular_clock (alias: dvmc)
                    - reports the degree of violation of the molecular clock
                evolutionary_rate (alias: evo_rate)
                    - reports a tree-based estimation of evolutionary rate for a gene
                internal_branch_stats (alias: ibs)
                    - calculates summary statistics for internal branch lengths 
                internode_labeler (alias: il)
                    - create labels at internodes in a phylogeny
                last_common_ancestor_subtree (alias: lca_subtree)
                    - get last common ancestor of a set of taxa
                long_branch_score (alias: lb_score; lbs)
                    - calculates lb (long branch) score for taxa in a phylogeny
                total_tree_length (alias: tree_len)
                    - calculates total tree length
                patristic_distances (alias: pd)
                    - calculate all pairwise distances between tips in a tree
                polytomy_test (alias: polyt_test; polyt; ptt)
                    - conducts a polytomy test using gene
                      support frequencies
                print_tree (alias: print; pt)
                    - prints ascii tree
                prune_tree (alias: prune)
                    - prune taxa from a phylogeny
                rename_tree_tips (alias: rename_tree; rename_tips)
                    - renames tips in a phylogeny according to a file with
                      the desired new tip names
                robinson_foulds_distance (alias: rf_distance; rf_dist; rf)
                    - calculates Robinson-Foulds distance between two trees
                spurious_sequence (alias: spurious_seq; ss)
                    - identifies putatively spurious sequences by identifying
                      branch lengths that are atypically long
                tip_labels (alias: tree_labels; labels; tl)
                    - print leaf names in a phylogeny
                treeness (alias: tness)
                    - reports treeness or stemminess, a measure of signal-to-
                      noise ratio in a phylogeny
             
                Alignment- and tree-based commands
                ==================================
                saturation (alias: sat)
                    - calculates saturation by examining the slope of
                      patristic distance and uncorrected distances
                treeness_over_rcv (alias: toverr)
                    - calculates treeness/rcv, treeness, and rcv
                """
            ),
        )
        parser.add_argument("command", help=SUPPRESS)
        args = parser.parse_args(sys.argv[1:2])

        # if command is part of the possible commands (i.e., the long form
        # commands, run). Otherwise, assume it is an alias and look to the
        # run_alias function
        try:
            if hasattr(self, args.command):
                getattr(self, args.command)(sys.argv[2:])
            else:
                self.run_alias(args.command, sys.argv[2:])
        except NameError:
            sys.exit()

    ## Aliases
    def run_alias(self, command, argv):
        # version
        if command in ['v']:
            return self.version()
        # Alignment aliases
        if command in ['aln_len', 'al']:
            return self.alignment_length(argv)
        elif command in ['aln_len_no_gaps', 'alng']:
            return self.alignment_length_no_gaps(argv)
        elif command in 'cs':
            return self.column_score(argv)
        elif command in ['get_entry', 'ge']:
            return self.faidx(argv)
        elif command == 'gc':
            return self.gc_content(argv)
        elif command in ['pairwise_id', 'pi']:
            return self.pairwise_identity(argv)
        elif command == 'pis':
            return self.parsimony_informative_sites(argv)
        elif command in ['rel_comp_var', 'relative_composition_variability']:
            return self.rcv(argv)
        elif command == 'rename_fasta':
            return self.rename_fasta_entries(argv)
        elif command in ['sum_of_pairs_score', 'sops', 'sop']:
            return self.sum_of_pairs_score(argv)
        elif command == 'vs':
            return self.variable_sites(argv)
        # Tree aliases
        elif command == 'bss':
            return self.bipartition_support_stats(argv)
        elif command == 'blm':
            return self.branch_length_multiplier(argv)
        elif command in ['collapse', 'cb']:
            return self.collapse_branches(argv)
        elif command == 'cover':
            return self.covarying_evolutionary_rates(argv)
        elif command == 'degree_of_violation_of_a_molecular_clock':
            return self.dvmc(argv)
        elif command == 'evo_rate':
            return self.evolutionary_rate(argv)
        elif command == 'ibs':
            return self.internal_branch_stats(argv)
        elif command == 'il':
            return self.internode_labeler(argv)
        elif command in ['lca_subtree']:
            return self.last_common_ancestor_subtree(argv)
        elif command in ['long_branch_score', 'lbs']:
            return self.lb_score(argv)
        elif command == 'pd':
            return self.patristic_distances(argv)
        elif command in ['polyt_test', 'ptt', 'polyt']:
            return self.polytomy_test(argv)
        elif command in ['print', 'pt']:
            return self.print_tree(argv)
        elif command == 'prune':
            return self.prune_tree(argv)
        elif command in ['rename_tree', 'rename_tips']:
            return self.rename_tree_tips(argv)
        elif command in ['robinson_foulds_distance', 'rf_dist', 'rf']:
            return self.rf_distance(argv)
        elif command in ['spurious_seq', 'ss']:
            return self.spurious_sequence(argv)
        elif command in ['labels', 'tree_labels', 'tl']:
            return self.tip_labels(argv)
        elif command == 'tree_len':
            return self.total_tree_length(argv)
        elif command == 'tness':
            return self.treeness(argv)
        # Alignment- and tree-based aliases
        elif command == 'sat':
            return self.saturation(argv)
        elif command in ['toverr', 'tor']:
            return self.treeness_over_rcv(argv)
        # Helper aliases
        elif command in ['create_concat', 'cc']:
            return self.create_concatenation_matrix(argv)
        elif command in ['pal2nal', 'p2n']:
            return self.thread_dna(argv)
        else:
            print("Invalid command option. See help for a complete list of commands and aliases.")
            parser.print_help()
            sys.exit(1)

    ## print version
    def version(self):
        print(textwrap.dedent(
            f"""\
            {self.help_header}
            """
        ))

    ## Alignment functions
    @staticmethod
    def alignment_length(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Length of an input alignment is calculated using this function.

                Longer alignments are associated with strong phylogenetic signal.
                
                Association between alignment length and phylogenetic signal
                was determined by Shen et al., Genome Biology and Evolution (2016),
                doi: 10.1093/gbe/evw179.

                Aliases:
                  alignment_length, aln_len, al
                Command line interfaces:
                  pk_alignment_length, pk_aln_len, pk_al

                Usage:
                phykit alignment_length <alignment>

                Options
                =====================================================
                <alignment>                 first argument after 
                                            function name should be
                                            an alignment file           
                """
            ),
        )
        parser.add_argument("alignment", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        AlignmentLength(args).run()

    @staticmethod
    def alignment_length_no_gaps(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate alignment length excluding sites with gaps.

                Longer alignments when excluding sites with gaps is
                associated with strong phylogenetic signal.

                PhyKIT reports three tab delimited values:
                col1: number of sites without gaps
                col2: total number of sites
                col3: percentage of sites without gaps

                Association between alignment length when excluding sites
                with gaps and phylogenetic signal was determined by Shen 
                et al., Genome Biology and Evolution (2016), 
                doi: 10.1093/gbe/evw179.

                Aliases:
                  alignment_length_no_gaps, aln_len_no_gaps, alng
                Command line interfaces: 
                  pk_alignment_length_no_gaps, pk_aln_len_no_gaps, pk_alng

                Usage:
                phykit alignment_length_no_gaps <alignment>

                Options
                =====================================================
                <alignment>                 first argument after 
                                            function name should be
                                            an alignment file          
                """
            ),
        )
        parser.add_argument("alignment", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        AlignmentLengthNoGaps(args).run()

    @staticmethod
    def column_score(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculates column score.

                Column is an accuracy metric for a multiple alignment relative
                to a reference alignment. It is calculated by summing the correctly
                aligned columns over all columns in an alignment. Thus, values range
                from 0 to 1 and higher values indicate more accurate alignments.

                Aliases:
                  column_score, cs
                Command line interfaces: 
                  pk_column_score, pk_cs

                Usage:
                phykit column_score <fasta> -r/--reference <ref.aln>

                Options
                =====================================================
                <fasta>                     first argument after 
                                            function name should be a
                                            query fasta alignment file
                                            to be scored for accuracy

                -r/--reference              reference fasta alignment to 
                                            compare query alignment to
                """
            ),
        )
        parser.add_argument("fasta", type=str, help=SUPPRESS)
        parser.add_argument("-r","--reference", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        ColumnScore(args).run()

    @staticmethod
    def faidx(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Extracts sequence entry from fasta file.

                This function works similarly to the faidx function 
                in samtools, but does not requiring an indexing function.

                Aliases:
                  faidx, get_entry; ge
                Command line interfaces: 
                  pk_faidx, pk_get_entry, pk_ge
                  

                Usage:
                phykit faidx <fasta> -e/--entry <fasta entry>

                Options
                =====================================================
                <fasta>                     first argument after 
                                            function name should be a
                                            query fasta file

                -e/--entry                  entry name to be extracted
                                            from the inputted fasta file
                """
            ),
        )
        parser.add_argument("fasta", type=str, help=SUPPRESS)
        parser.add_argument("-e","--entry", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        Faidx(args).run()

    @staticmethod
    def gc_content(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}
                
                Calculate GC content of a fasta file.

                GC content is negatively correlated with phylogenetic signal.

                If there are multiple entries, use the -v/--verbose option
                to determine the GC content of each fasta entry separately.

                Association between GC content and phylogenetic signal was
                determined by Shen et al., Genome Biology and Evolution (2016), 
                doi: 10.1093/gbe/evw179.

                Aliases:
                  gc_content, gc
                Command line interfaces: 
                  pk_gc_content, pk_gc

                Usage:
                phykit gc_content <fasta> [-v/--verbose]

                Options
                =====================================================
                <fasta>                     first argument after 
                                            function name should be
                                            a fasta file 
            
                -v, --verbose               optional argument to print
                                            the GC content of each fasta
                                            entry
                """
            ),
        )
        parser.add_argument("fasta", type=str, help=SUPPRESS)
        parser.add_argument("-v", "--verbose", action="store_true", required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        GCContent(args).run()

    @staticmethod
    def pairwise_identity(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate the average pairwise identity among sequences.
                
                Pairwise identities can be used as proxies for the
                evolutionary rate of sequences.

                Pairwise identity is defined as the number of identical
                columns (including gaps) between two aligned sequences divided
                by the number of columns in the alignment. Summary statistics
                are reported unless used with the verbose option in which
                all pairwise identities will be reported.

                An example of pairwise identities being used as a proxy
                for evolutionary rate can be found here: Chen et al. 
                Genome Biology and Evolution (2017), doi: 10.1093/gbe/evx147.

                Aliases:
                  pairwise_identity, pairwise_id, pi
                Command line interfaces:
                  pk_pairwise_identity, pk_pairwise_id, pk_pi

                Usage:
                phykit pairwise_identity <alignment> [-v/--verbose]

                Options
                =====================================================
                <alignment>                 first argument after 
                                            function name should be
                                            an alignment file  

                -v, --verbose               optional argument to print
                                            identity per pair       
                """
            ),
        )
        parser.add_argument("alignment", type=str, help=SUPPRESS)
        parser.add_argument("-v", "--verbose", action="store_true", required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        PairwiseIdentity(args).run()

    @staticmethod
    def parsimony_informative_sites(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate the number and percentage of parismony
                informative sites in an alignment.

                The number of parsimony informative sites in an alignment
                is associated with strong phylogenetic signal.

                PhyKIT reports three tab delimited values:
                col1: number of parsimony informative sites
                col2: total number of sites
                col3: percentage of parsimony informative sites

                Association between the number of parsimony informative
                sites and phylogenetic signal was determined by Shen 
                et al., Genome Biology and Evolution (2016), 
                doi: 10.1093/gbe/evw179 and Steenwyk et al., bioRxiv
                (2020), doi: 10.1101/2020.06.08.140384.

                Aliases: 
                  parsimony_informative_sites, pis
                Command line interfaces:
                  pk_parsimony_informative_sites, pk_pis

                Usage:
                phykit parsimony_informative_sites <alignment>

                Options
                =====================================================
                <alignment>                 first argument after 
                                            function name should be
                                            an alignment file          
                """
            ),
        )
        parser.add_argument("alignment", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        ParsimonyInformative(args).run()

    @staticmethod
    def rcv(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate RCV (relative composition variability) for an alignment.

                Lower RCV values are thought to be desirable because they represent
                a lower composition bias in an alignment. Statistically, RCV describes
                the average variability in sequence composition among taxa. 

                RCV is calculated following Phillips and Penny, Molecular Phylogenetics
                and Evolution (2003), doi: 10.1016/S1055-7903(03)00057-5.

                Aliases: 
                  relative_composition_variability, rel_comp_var, rcv
                Command line interfaces:
                  pk_relative_composition_variability, pk_rel_comp_var, pk_rcv

                Usage:
                phykit relative_composition_variability <alignment>

                Options
                =====================================================
                <alignment>                 first argument after 
                                            function name should be
                                            an alignment file          
                """
            ),
        )
        parser.add_argument("alignment", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        RelativeCompositionVariability(args).run()

    @staticmethod
    def rename_fasta_entries(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Renames fasta entries.

                Renaming fasta entries will follow the scheme of a tab-delimited
                file wherein the first column is the current fasta entry name and
                the second column is the new fasta entry name in the resulting 
                output alignment. 

                Aliases:
                  rename_fasta_entries, rename_fasta
                Command line interfaces: 
                  pk_rename_fasta_entries, pk_rename_fasta

                Usage:
                phykit rename_fasta_entries <fasta> -i/--idmap <idmap>
                    [-o/--output <output_file>]

                Options
                =====================================================
                <fasta>                     first argument after 
                                            function name should be
                                            a fasta file

                -i/--idmap                  identifier map of current FASTA
                                            names (col1) and desired FASTA
                                            names (col2)

                -o/--output                 optional argument to write
                                            the renamed fasta file to.
                                            Default output has the same 
                                            name as the input file with
                                            the suffix ".renamed.fa" added
                                            to it.
                """
            ),
        )
        parser.add_argument("fasta", type=str, help=SUPPRESS)
        parser.add_argument("-i","--idmap", type=str, help=SUPPRESS)
        parser.add_argument("-o", "--output", type=str, required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        RenameFastaEntries(args).run()

    @staticmethod
    def sum_of_pairs_score(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculates sum-of-pairs score.

                Sum-of-pairs is an accuracy metric for a multiple alignment relative
                to a reference alignment. It is calculated by summing the correctly
                aligned residue pairs over all pairs of sequences. Thus, values range
                from 0 to 1 and higher values indicate more accurate alignments.

                Aliases:
                  sum_of_pairs_score, sops, sop 
                Command line interfaces: 
                  pk_sum_of_pairs_score, pk_sops, pk_sop

                Usage:
                phykit sum_of_pairs_score <fasta> -r/--reference <ref.aln>

                Options
                =====================================================
                <fasta>                     first argument after 
                                            function name should be a
                                            query fasta alignment file
                                            to be scored for accuracy

                -r/--reference              reference fasta alignment to 
                                            compare query alignment to
                """
            ),
        )
        parser.add_argument("fasta", type=str, help=SUPPRESS)
        parser.add_argument("-r","--reference", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        SumOfPairsScore(args).run()

    @staticmethod
    def variable_sites(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate the number of variable sites in an alignment.

                The number of variable sites in an alignment is 
                associated with strong phylogenetic signal.

                PhyKIT reports three tab delimited values:
                col1: number of variable sites
                col2: total number of sites
                col3: percentage of variable sites

                Association between the number of variable sites and
                phylogenetic signal was determined by Shen et al.,
                Genome Biology and Evolution (2016), 
                doi: 10.1093/gbe/evw179.

                Aliases:
                  variable_sites, vs
                Command line interfaces: 
                  pk_variable_sites, pk_vs

                Usage:
                phykit variable_sites <alignment>

                Options
                =====================================================
                <alignment>                 first argument after 
                                            function name should be
                                            an alignment file          
                """
            ),
        )
        parser.add_argument("alignment", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        VariableSites(args).run()


    ## Tree functions
    @staticmethod
    def bipartition_support_stats(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}
                Calculate summary statistics for bipartition support.

                High bipartition support values are thought to be desirable because
                they are indicative of greater certainty in tree topology.

                To obtain all bipartition support values, use the -v/--verbose option.

                Aliases:
                  bipartition_support_stats, bss
                Command line interfaces:
                  pk_bipartition_support_stats, pk_bss

                Usage:
                phykit bipartition_support_stats <tree> [-v/--verbose]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file 
            
                -v, --verbose               optional argument to print
                                            all bipartition support
                                            values
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            required=False,
            help=SUPPRESS
        )
        args = parser.parse_args(argv)
        BipartitionSupportStats(args).run()

    @staticmethod
    def branch_length_multiplier(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header} 

                Multiply branch lengths in a phylogeny by a given factor.
                
                This can help modify reference trees when conducting simulations
                or other analyses.              

                Alias:
                  branch_length_multiplier, blm
                Command line interfaces:
                  pk_branch_length_multiplier, pk_blm

                Usage:
                phykit branch_length_multiplier <tree> -f n [-o/--output <output_file>]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            an tree file

                -f/--factor                 factor to multiply branch 
                                            lengths by 

                -o/--output                 optional argument to name 
                                            the outputted tree file.
                                            Default output will have 
                                            the same name as the input
                                            file but with the suffix 
                                            ".factor_(n).tre"
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument(
            "-f", "--factor",
            type=float, required=True,
            help=SUPPRESS
        )
        parser.add_argument("-o", "--output", type=str, required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        BranchLengthMultiplier(args).run()

    @staticmethod
    def collapse_branches(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header} 

                Collapse branches on a phylogeny according to bipartition support.
                Bipartitions will be collapsed if they are less than the user specified
                value.              

                Aliases:
                  collapse_branches, collapse, cb
                Command line interfaces:
                  pk_collapse_branches, pk_collapse, pk_cb

                Usage:
                phykit collapse_branches <tree> -s/--support n [-o/--output <output_file>]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            an tree file

                -s/--support                bipartitions with support less
                                            than this value will be collapsed

                -o/--output                 optional argument to name 
                                            the outputted tree file.
                                            Default output will have 
                                            the same name as the input
                                            file but with the suffix 
                                            ".collapsed_(support).tre"

                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument(
            "-s", "--support",
            type=float, required=True,
            help=SUPPRESS
        )
        parser.add_argument("-o", "--output", type=str, required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        CollapseBranches(args).run()

    @staticmethod
    def covarying_evolutionary_rates(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Determine if two genes have a signature of covariation with one another.

                Genes that have covarying evolutionary histories tend to have 
                similar functions and expression levels.

                Input two phylogenies and calculate the correlation among relative 
                evolutionary rates between the two phylogenies. The two input trees 
                do not have to have the same taxa. This function will first prune both
                trees to have the same tips. To transform branch lengths into relative
                rates, PhyKIT uses the putative species tree's branch lengths, which is
                inputted by the user. As recommended by the original method developers,
                outlier branche lengths are removed. Outlier branches have a relative 
                evolutionary rate greater than five.

                PhyKIT reports two tab delimited values:
                col1: correlation coefficient
                col2: p-value

                Method is empirically evaluated by Clark et al., Genome Research
                (2012), doi: 10.1101/gr.132647.111. Normalization method using a 
                species tree follows Sato et al., Bioinformatics (2005), doi: 
                10.1093/bioinformatics/bti564. 


                Aliases:
                  covarying_evolutionary_rates, cover
                Command line interfaces:
                  pk_covarying_evolutionary_rates, pk_cover

                Usage:
                phykit covarying_evolutionary_rates <tree_file_zero> <tree_file_one>
                    -r/--reference <reference_tree_file> [-v/--verbose]

                Options
                =====================================================
                <tree_file_zero>            first argument after 
                                            function name should be
                                            an alignment file

                <tree_file_one>             first argument after 
                                            function name should be
                                            an alignment file 

                -r/--reference              a tree to correct branch
                                            lengths by in the two input
                                            trees. Typically, this is a
                                            putative species tree.

                -v/--verbose                print out corrected branch
                                            lengths shared between
                                            tree 0 and tree 1
                """
            ),
        )
        parser.add_argument("tree_zero", type=str, help=SUPPRESS)
        parser.add_argument("tree_one", type=str, help=SUPPRESS)
        parser.add_argument(
            "-r", "--reference", type=str, required=True, help=SUPPRESS, metavar=""
        )
        parser.add_argument("-v", "--verbose", action="store_true", required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        CovaryingEvolutionaryRates(args).run()

    @staticmethod
    def dvmc(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate degree of violation of a molecular clock (or DVMC) in a phylogeny.

                Lower DVMC values are thought to be desirable because they are indicative
                of a lower degree of violation in the molecular clock assumption.

                Typically, outgroup taxa are not included in molecular clock analysis. Thus,
                prior to calculating DVMC from a single gene tree, outgroup taxa are pruned
                from the phylogeny. PhyKIT will prune taxa using tip names from a single column
                file, which is specified using the -r/--root file. If the tip name does not
                exist in the input tree, rather than raising an error/warning message, the tip  
                name is skipped. If the user wants to calculate DVMC for phylogenies with incomplete
                taxa representation, this will allow the user to use one <root file> for all trees.
                Lastly, an empty root file can be used if the user does not wish to prune outgroup taxa. 

                Calculate DVMC in a tree following Liu et al., PNAS (2017), doi: 10.1073/pnas.1616744114.

                Aliases:
                  degree_of_violation_of_a_molecular_clock, dvmc
                Command line interfaces:
                  pk_degree_of_violation_of_a_molecular_clock, pk_dvmc

                Usage:
                phykit degree_of_violation_of_a_molecular_clock -t/--tree <tree>
                    -r/--root <root_taxa>

                Options
                =====================================================
                -t, --tree                  input file tree name
            
                -r, --root                  single column file with
                                            tip names of root taxa
                """
            ),
        )
        parser.add_argument(
            "-r", "--root", type=str, required=True, help=SUPPRESS, metavar=""
        )
        parser.add_argument(
            "-t", "--tree", type=str, required=True, help=SUPPRESS, metavar=""
        )
        args = parser.parse_args(argv)
        DVMC(args).run()

    @staticmethod
    def evolutionary_rate(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}
                Calculate a tree-based estimation of the evolutionary rate of a gene.

                Evolutionary rate is the total tree length divided by the number
                of terminals.

                Calculate evolutionary rate following Telford et al., Proceedings
                of the Royal Society B (2014).

                Aliases:
                  evolutionary_rate, evo_rate
                Command line interfaces:
                  pk_evolutionary_rate, pk_evo_rate

                Usage:
                phykit evolutionary_rate <tree>

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file 
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        EvolutionaryRate(args).run()

    @staticmethod
    def internal_branch_stats(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate summary statistics for internal branch lengths in a phylogeny.

                Internal branch lengths can be useful for phylogeny diagnostics.

                To obtain all internal branch lengths, use the -v/--verbose option. 

                Aliases:
                  internal_branch_stats, ibs
                Command line interfaces:
                  pk_internal_branch_stats, pk_ibs

                Usage:
                phykit internal_branch_stats <tree> [-v/--verbose]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file

                -v, --verbose               optional argument to print
                                            all internal branch lengths
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            required=False,
            help=SUPPRESS
        )
        args = parser.parse_args(argv)
        InternalBranchStats(args).run()

    @staticmethod
    def internode_labeler(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Appends numerical identifiers to bipartitions in place
                of support values. This is helpful for pointing to
                specific internodes in supplementary files or otherwise. 

                Alias:
                  internode_labeler, il
                Command line interfaces: 
                  pk_internode_labeler, pk_il

                Usage:
                phykit internode_labeler <file> [-o/--output <file>]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file

                -o/--output                 optional argument to name 
                                            the outputted tree file
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument("-o", "--output", type=str, required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        InternodeLabeler(args).run()

    @staticmethod
    def last_common_ancestor_subtree(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Obtains subtree from a phylogeny by getting
                the last common ancestor from a list of taxa. 

                Alias:
                  last_common_ancestor_subtree, lca_subtree
                Command line interfaces: 
                  pk_last_common_ancestor_subtree, pk_lca_subtree

                Usage:
                phykit last_common_ancestor_subtree <file> <list_of_taxa> 
                [-o/--output <file>]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file

                <list_of_taxa>              list of taxa to get the last
                                            common ancestor subtree for

                -o/--output                 optional argument to name 
                                            the outputted tree file
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument("list_of_taxa", type=str, help=SUPPRESS)
        parser.add_argument("-o", "--output", type=str, required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        LastCommonAncestorSubtree(args).run()

    @staticmethod
    def lb_score(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate long branch (LB) scores in a phylogeny.

                Lower LB scores are thought to be desirable because
                they are indicative of taxa or trees that likely do
                not have issues with long branch attraction.

                LB score is the mean pairwise patristic distance of
                taxon i compared to all other taxa over the average 
                pairwise patristic distance. 
        
                PhyKIT reports summary statistics. To obtain LB scores
                for each taxa, use the -v/--verbose option. 

                LB scores are calculated following Struck, Evolutionary 
                Bioinformatics (2014), doi: 10.4137/EBO.S14239.

                Aliases:
                  long_branch_score, lb_score, lbs
                Command line interfaces:
                  pk_long_branch_score, pk_lb_score, pk_lbs

                Usage:
                phykit long_branch_score <tree> [-v/--verbose]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file

                -v, --verbose               optional argument to print
                                            all LB score values            
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument("-v", "--verbose", action="store_true", required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        LBScore(args).run()

    @staticmethod
    def patristic_distances(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate summary statistics among patristic distances in a phylogeny.

                Patristic distances are all tip-to-tip distances in a phylogeny.

                To obtain all patristic distances, use the -v/--verbose option.
                With the -v option, the first column will have two taxon names
                separated by a '-' followed by the patristic distance. Features
                will be tab separated. 

                Aliases:
                  patristic_distances, pd
                Command line interfaces: 
                  pk_patristic_distances, pk_pd

                Usage:
                phykit patristic_distances <tree> [-v/--verbose]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file

                -v, --verbose               optional argument to print
                                            all patristic distances between
                                            taxa            
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument("-v", "--verbose", action="store_true", required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        PatristicDistances(args).run() 

    @staticmethod
    def polytomy_test(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Conduct a polytomy test for three clades in a phylogeny.

                Polytomy tests can be used to identify putative radiations
                as well as identify well supported alternative topologies.

                The polytomy testing function takes as input a file with
                the three groups of taxa to test the relationships for and
                a single column file with the names of the desired tree files
                to use for polytomy testing. Next, the script to examine
                support for the grouping of the three taxa using triplets
                and gene support frequencies. 

                This function can account for uncertainty in gene trees - 
                that is, the input phylogenies can have collapsed bipartitions.

                Thereafter, a chi-squared test is conducted to determine if there
                is evidence to reject the null hypothesis wherein the null 
                hypothesis is that the three possible topologies among the three
                groups are equally supported. This test is done using gene support
                frequencies.

                Aliases:
                  polytomy_test, polyt_test, polyt, ptt
                Command line interfaces: 
                  pk_polytomy_test, pk_polyt_test, pk_polyt, pk_ptt

                Usage:
                phykit polytomy_test -t/--trees <trees> -g/--groups <groups>

                Options
                =====================================================
                -t, --trees                 single column file with names
                                            of phylogenies to use for
                                            polytomy testing

                -g, --groups                a tab-delimited file with the
                                            grouping designations to test.
                                            Lines starting with comments 
                                            are not considered. Names
                                            of individual taxa should be
                                            separated by a semi-colon  ';'
                                            
                For example, the groups file could look like the following:
                #labels group0  group1  group2
                name_of_test    tip_name_A;tip_name_B   tip_name_C  tip_name_D;tip_name_E
                """
            ),
        )
        parser.add_argument("-t", "--trees", type=str, help=SUPPRESS)
        parser.add_argument("-g", "--groups", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        PolytomyTest(args).run() 

    @staticmethod
    def print_tree(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Print ascii tree of input phylogeny.

                Phylogeny can be printed with or without branch lengths.
                By default, the phylogeny will be printed with branch lengths
                but branch lengths can be removed using the -r/--remove argument.

                Aliases:
                  print_tree, print, pt
                Command line interfaces:
                  pk_print_tree, pk_print, pk_pt

                Usage:
                phykit print_tree <tree> [-r/--remove]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file

                -r, --remove                optional argument to print
                                            the phylogeny without branch
                                            lengths       
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument("-r", "--remove", action="store_true", required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        PrintTree(args).run()

    @staticmethod
    def prune_tree(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Prune tips from a phylogeny.

                Provide a single column file with the names of the tips
                in the input phylogeny you would like to prune from the
                tree.

                Aliases: 
                  prune_tree, prune
                Command line interfaces: 
                  pk_prune_tree, pk_prune

                Usage:
                phykit prune_tree <tree> <list_of_taxa> [-o/--output <output_file>]

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file

                <list_of_taxa>              single column file with the
                                            names of the tips to remove
                                            from the phylogeny 

                -o/--output                 name of output file for the
                                            pruned phylogeny. 
                                            Default output will have 
                                            the same name as the input
                                            file but with the suffix 
                                            ".pruned"    
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument("list_of_taxa", type=str, help=SUPPRESS)
        parser.add_argument("-o", "--output", type=str, required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        PruneTree(args).run()

    @staticmethod
    def rename_tree_tips(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Renames tips in a phylogeny.

                Renaming tip files will follow the scheme of a tab-delimited
                file wherein the first column is the current tip name and the
                second column is the desired tip name in the resulting 
                phylogeny. 

                Aliases:
                  rename_tree_tips, rename_tree, rename_tips
                Command line interfaces: 
                  pk_rename_tree_tips, pk_rename_tree, pk_rename_tips

                Usage:
                phykit rename_tree_tips <tree> -i/--idmap <idmap.txt>
                    [-o/--output <output_file>] 

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file

                -i/--idmap                  identifier map of current tip
                                            names (col1) and desired tip
                                            names (col2)

                -o/--output                 optional argument to write
                                            the renamed tree files to.
                                            Default output will have 
                                            the same name as the input
                                            file but with the suffix 
                                            ".renamed"          
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument("-i","--idmap", type=str, help=SUPPRESS)
        parser.add_argument("-o", "--output", type=str, required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        RenameTreeTips(args).run()

    @staticmethod
    def rf_distance(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate Robinson-Foulds (RF) distance between two trees.

                Low RF distances reflect greater similarity between two phylogenies. 
                This function prints out two values, the plain RF value and the
                normalized RF value, which are separated by a tab. Normalized RF values
                are calculated by taking the plain RF value and dividing it by 2(n-3)
                where n is the number of tips in the phylogeny. Prior to calculating
                an RF value, PhyKIT will first determine the number of shared tips
                between the two input phylogenies and prune them to a common set of
                tips. Thus, users can input trees with different topologies and 
                infer an RF value among subtrees with shared tips.

                PhyKIT will print out 
                col 1; the plain RF distance and 
                col 2: the normalized RF distance.

                RF distances are calculated following Robinson & Foulds, Mathematical 
                Biosciences (1981), doi: 10.1016/0025-5564(81)90043-2.

                Aliases:
                  robinson_foulds_distance, rf_distance, rf_dist, rf
                Command line interfaces: 
                  pk_robinson_foulds_distance, pk_rf_distance, pk_rf_dist, pk_rf

                Usage:
                phykit robinson_foulds_distance <tree_file_zero> <tree_file_one>

                Options
                =====================================================
                <tree_file_zero>            first argument after 
                                            function name should be
                                            a tree file

                <tree_file_one>             second argument after 
                                            function name should be
                                            a tree file           
                """
            ),
        )
        parser.add_argument("tree_zero", type=str, help=SUPPRESS)
        parser.add_argument("tree_one", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        RobinsonFouldsDistance(args).run()

    @staticmethod
    def spurious_sequence(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header} 

                Determines potentially spurious homologs using branch lengths.

                Identifies potentially spurious sequences and reports
                tips in the phylogeny that could possibly be removed
                from the associated multiple sequence alignment. PhyKIT
                does so by identifying and reporting long terminal branches
                defined as branches that are equal to or 20 times the median
                length of all branches.

                PhyKIT reports the following information
                col1: name of tip that is a putatively spurious sequence
                col2: length of branch leading to putatively spurious sequence
                col3: threshold used to identify putatively spurious sequences
                col4: median branch length in the phylogeny

                If there are no putatively spurious sequences, "None" is reported.
                
                Using this method to identify potentially spurious sequences
                was, to my knowledge, first introduced by Shen et al., (2018)
                Cell doi: 10.1016/j.cell.2018.10.023.                

                Aliases:
                  spurious_sequence, spurious_seq, ss
                Command line interfaces:
                  pk_spurious_sequence, pk_spurious_seq, pk_ss

                Usage:
                phykit spurious_sequence <file> [-f 20]

                Options
                =====================================================
                <file>                      first argument after 
                                            function name should be
                                            an tree file

                -f/--factor                 factor to multiply median
                                            branch length by to calculate
                                            the threshold of long branches.
                                            (Default: 20)
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        parser.add_argument(
            "-f", "--factor",
            type=float, required=False,
            help=SUPPRESS
        )
        args = parser.parse_args(argv)
        SpuriousSequence(args).run()

    @staticmethod
    def tip_labels(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Prints the tip labels (or names) a phylogeny.

                Aliases:
                  tip_labels, tree_labels; labels; tl
                Command line interfaces: 
                  pk_tip_labels, pk_tree_labels; pk_labels; pk_tl

                Usage:
                phykit tip_labels <tree>

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        TipLabels(args).run()

    @staticmethod
    def total_tree_length(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate total tree length, which is a sum of all branches. 

                Aliases:
                  total_tree_length, tree_len
                Command line interfaces: 
                  pk_total_tree_length, pk_tree_len

                Usage:
                phykit total_tree_length <tree>

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        TotalTreeLength(args).run()

    @staticmethod
    def treeness(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate treeness statistic for a phylogeny.

                Higher treeness values are thought to be desirable because they
                represent a higher signal-to-noise ratio.

                Treeness describes the proportion of the tree distance found on
                internal branches. Treeness can be used as a measure of the 
                signal-to-noise ratio in a phylogeny. 

                Calculate treeness (also referred to as stemminess) following
                Lanyon, The Auk (1988), doi: 10.1093/auk/105.3.565 and
                Phillips and Penny, Molecular Phylogenetics and Evolution
                (2003), doi: 10.1016/S1055-7903(03)00057-5.

                Aliases:
                  treeness, tness
                Command line interfaces:
                  pk_treeness, pk_tness

                Usage:
                phykit treeness <tree>

                Options
                =====================================================
                <tree>                      first argument after 
                                            function name should be
                                            a tree file
                """
            ),
        )
        parser.add_argument("tree", type=str, help=SUPPRESS)
        args = parser.parse_args(argv)
        Treeness(args).run()

    ## Alignment and tree functions
    @staticmethod
    def saturation(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate saturation for a given tree and alignment.

                Saturation is defined as sequences in multiple sequence
                alignments that have undergone numerous substitutions such
                that the distances between taxa are underestimated.

                Data with no saturation will have a value of 1. Completely
                saturated data will have a value of 0.  

                Saturation is calculated following Philippe et al., PLoS 
                Biology (2011), doi: 10.1371/journal.pbio.1000602.

                Aliases: 
                  saturation, sat
                Command line interfaces:
                  pk_saturation, pk_sat

                Usage:
                phykit saturation -a <alignment> -t <tree> [-v/--verbose]

                Options
                =====================================================
                -a/--alignment              an alignment file
                
                -t/--tree                   a tree file

                -v/--verbose                print out patristic distances
                                            and uncorrected distances used
                                            to determine saturation
                """
            ),
        )
        parser.add_argument(
            "-a", "--alignment", type=str, required=True, help=SUPPRESS, metavar=""
        )
        parser.add_argument(
            "-t", "--tree", type=str, required=True, help=SUPPRESS, metavar=""
        )
        parser.add_argument("-v", "--verbose", action="store_true", required=False, help=SUPPRESS)
        args = parser.parse_args(argv)
        Saturation(args).run()

    @staticmethod
    def treeness_over_rcv(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Calculate treeness/RCV for a given alignment and tree.

                Higher treeness/RCV values are thought to be desirable because
                they harbor a high signal-to-noise ratio are least susceptible
                to composition bias.

                PhyKIT reports three tab delimited values:
                col1: treeness/RCV
                col2: treeness
                col3: RCV

                Calculate treeness/RCV following Phillips and Penny, Molecular 
                Phylogenetics and Evolution (2003), doi: 10.1016/S1055-7903(03)00057-5.

                Aliases:
                  treeness_over_rcv, toverr, tor
                Command line interfaces:
                  pk_treeness_over_rcv, pk_toverr, pk_tor

                Usage:
                phykit treeness_over_rcv -a/--alignment <alignment> -t/--tree <tree>

                Options
                =====================================================
                -a/--alignment              an alignment file
                
                -t/--tree                   a tree file
                """
            ),
        )
        parser.add_argument(
            "-a", "--alignment", type=str, required=True, help=SUPPRESS, metavar=""
        )
        parser.add_argument(
            "-t", "--tree", type=str, required=True, help=SUPPRESS, metavar=""
        )
        args = parser.parse_args(argv)
        TreenessOverRCV(args).run()

    ### Helper commands
    @staticmethod
    def create_concatenation_matrix(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Create a concatenated alignment file. This function is 
                used to help in the construction of multi-locus data
                matrices.

                PhyKIT will output three files:
                1) A fasta file with '.fa' appended to the prefix specified
                   with the -p/--prefix parameter.
                2) A partition file ready for input into RAxML or IQ-tree.
                3) An occupancy file that summarizes the taxon occupancy
                   per sequence.

                Aliases:
                  create_concatenation_matrix, create_concat, cc
                Command line interfaces: 
                  pk_create_concatenation_matrix, pk_create_concat, pk_cc

                Usage:
                phykit create_concatenation_matrix -a <file> -p <string>

                Options
                =====================================================
                -a/--alignment              alignment list file. File
                                            should contain a single
                                            column list of alignment
                                            sequence files to concatenate 
                                            into a single matrix. Provide
                                            path to files relative to
                                            working directory or provide
                                            absolute path.

                -p/--prefix                 prefix of output files
                """
            ),
        )
        parser.add_argument(
            "-a", "--alignment_list",
            type=str,
            help=SUPPRESS
        )
        parser.add_argument(
            "-p", "--prefix",
            type=str,
            help=SUPPRESS
        )
        args = parser.parse_args(argv)
        CreateConcatenationMatrix(args).run()

    @staticmethod
    def thread_dna(argv):
        parser = ArgumentParser(add_help=True,
            usage=SUPPRESS,
            formatter_class=RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                f"""\
                {help_header}

                Thread DNA sequence onto a protein alignment to create a
                codon-based alignment. 
                
                This function requires input alignments are in fasta format.
                Codon alignments are then printed to stdout. Note, sequences
                are assumed to occur in the same order in the protein and 
                nucleotide alignment.

                Aliases:
                  thread_dna, pal2nal, p2n
                Command line interfaces:
                  pk_thread_dna, pk_pal2nal, pk_p2n

                Usage:
                phykit thread_dna -p <file> -n <file> [-s]

                Options
                =====================================================
                -p/--protein                protein alignment file

                -n/--nucleotide             nucleotide alignment file

                -s/--stop                   boolean for whether or not
                                            stop codons should be kept. 
                                            If used, stop codons will 
                                            be removed.
                """
            ),
        )
        parser.add_argument(
            "-p", "--protein",
            type=str,
            help=SUPPRESS
        )
        parser.add_argument(
            "-n", "--nucleotide",
            type=str,
            help=SUPPRESS
        )
        parser.add_argument(
            "-s", "--stop",
            type=str2bool, 
            nargs='?',
            default=True,
            help=SUPPRESS
        )
        args = parser.parse_args(argv)
        DNAThreader(args).run()

def main(argv=None):
    Phykit()

# Alignment-based functions
def alignment_length(argv=None):
    Phykit.alignment_length(sys.argv[1:])

def alignment_length_no_gaps(argv=None):
    Phykit.alignment_length_no_gaps(sys.argv[1:])

def column_score(argv=None):
    Phykit.column_score(sys.argv[1:])

def faidx(argv=None):
    Phykit.faidx(sys.argv[1:])

def gc_content(argv=None):
    Phykit.gc_content(sys.argv[1:])

def pairwise_identity(argv=None):
    Phykit.pairwise_identity(sys.argv[1:])

def parsimony_informative_sites(argv=None):
    Phykit.parsimony_informative_sites(sys.argv[1:])

def rcv(argv=None):
    Phykit.rcv(sys.argv[1:])

def rename_fasta_entries(argv=None):
    Phykit.rename_fasta_entries(sys.argv[1:])

def sum_of_pairs_score(argv=None):
    Phykit.sum_of_pairs_score(sys.argv[1:])

def variable_sites(argv=None):
    Phykit.variable_sites(sys.argv[1:])

# Tree-based functions
def bipartition_support_stats(argv=None):
    Phykit.bipartition_support_stats(sys.argv[1:])

def branch_length_multiplier(argv=None):
    Phykit.branch_length_multiplier(sys.argv[1:])

def collapse_branches(argv=None):
    Phykit.collapse_branches(sys.argv[1:])

def covarying_evolutionary_rates(argv=None):
    Phykit.covarying_evolutionary_rates(sys.argv[1:])

def dvmc(argv=None):
    Phykit.dvmc(sys.argv[1:])

def evolutionary_rate(argv=None):
    Phykit.evolutionary_rate(sys.argv[1:])

def internal_branch_stats(argv=None):
    Phykit.internal_branch_stats(sys.argv[1:])

def internode_labeler(argv=None):
    Phykit.internode_labeler(sys.argv[1:])

def last_common_ancestor_subtree(argv=None):
    Phykit.last_common_ancestor_subtree(sys.argv[1:])

def lb_score(argv=None):
    Phykit.lb_score(sys.argv[1:])

def patristic_distances(argv=None):
    Phykit.patristic_distances(sys.argv[1:])

def polytomy_test(argv=None):
    Phykit.polytomy_test(sys.argv[1:])

def print_tree(argv=None):
    Phykit.print_tree(sys.argv[1:])

def prune_tree(argv=None):
    Phykit.prune_tree(sys.argv[1:])

def rename_tree_tips(argv=None):
    Phykit.rename_tree_tips(sys.argv[1:])

def rf_distance(argv=None):
    Phykit.rf_distance(sys.argv[1:])

def spurious_sequence(argv=None):
    Phykit.spurious_sequence(sys.argv[1:])

def tip_labels(argv=None):
    Phykit.tip_labels(sys.argv[1:])

def total_tree_length(argv=None):
    Phykit.total_tree_length(sys.argv[1:])

def treeness(argv=None):
    Phykit.treeness(sys.argv[1:])

# Alignment- and tree-based functions
def saturation(argv=None):
    Phykit.saturation(sys.argv[1:])

def treeness_over_rcv(argv=None):
    Phykit.treeness_over_rcv(sys.argv[1:])

# Helper functions
def create_concatenation_matrix(argv=None):
    Phykit.create_concatenation_matrix(sys.argv[1:])

def thread_dna(argv=None):
    Phykit.thread_dna(sys.argv[1:])




from enum import Enum

from Bio import AlignIO

from .base import Tree
from ..alignment.base import Alignment
from ...helpers.files import get_alignment_and_format

class FileFormat(Enum):
    fasta = "fasta"
    clustal = "clustal"
    maf = "maf"
    mauve = "mauve"
    phylip = "phylip"
    phylip_seq = "phylip-sequential"
    phylip_rel = "phylip-relaxed"
    stockholm = "stockholm"

class TreenessOverRCV(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # calculate treeness
        treeness = self.calculate_treeness()
        
        # calculate rcv
        aln = Alignment(alignment_file_path=self.alignment_file_path)
        relative_composition_variability = aln.calculate_rcv()

        # calculate treeness/rcv
        treeness_over_rcv = self.calculate_treeness_over_rcv(treeness, relative_composition_variability)
        
        # print results
        print(f"{round(treeness_over_rcv, 4)}\t{round(treeness, 4)}\t{round(relative_composition_variability, 4)}")


    def process_args(self, args):
        return dict(tree_file_path=args.tree, alignment_file_path=args.alignment)
    
    def calculate_treeness_over_rcv(self, treeness, relative_composition_variability):
        treeness_over_rcv = (treeness / relative_composition_variability)
        return treeness_over_rcv




from Bio import SeqIO
from Bio.Seq import Seq

from .base import Alignment

class RelativeCompositionVariability(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # calc rcv and print val
        relative_composition_variability = self.calculate_rcv()
        print(round(relative_composition_variability, 4))


    def process_args(self, args):
        return dict(alignment_file_path=args.alignment)


from collections import Counter
import itertools
import sys

from Bio import SeqIO, AlignIO

from .base import Alignment

class ColumnScore(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # create biopython object of sequences for query and 
        # reference alignments
        query_records = AlignIO.read(self.fasta, "fasta")
        reference_records = AlignIO.read(self.reference, "fasta")

        # create lists with strings of every columns
        ref_columns, query_columns = self.get_columns_from_alignments(reference_records, query_records)
        
        # count the number of matches and total pairs
        number_of_matches, number_of_total_columns = self.calculate_matches_between_ref_and_query_columns(ref_columns, query_columns)

        # print res
        print(round(number_of_matches/number_of_total_columns, 4))

    def process_args(self, args) -> dict:
        return dict(
            fasta=args.fasta,
            reference=args.reference
        )

    def get_columns_from_alignments(
        self,
        reference_records,
        query_records
    ):
        # loop through reference alignment and get each column
        ref_columns = []
        for i in range(reference_records.get_alignment_length()):
            ref_seq_at_position = ''
            ref_seq_at_position += reference_records[:,i]
            ref_seq_at_position = ref_seq_at_position.upper()
            ref_columns.append(ref_seq_at_position)

        # loop through query alignment and get each column
        query_columns = []
        for i in range(query_records.get_alignment_length()):
            query_seq_at_position = ''
            query_seq_at_position += query_records[:,i]
            query_seq_at_position = query_seq_at_position.upper()
            query_columns.append(query_seq_at_position)

        return ref_columns, query_columns

    def calculate_matches_between_ref_and_query_columns(
        self,
        ref_columns: list,
        query_columns :list
    ):
        # determine the number of matches
        matches = list((Counter(ref_columns) & Counter(query_columns)).elements())
        return len(matches), len(ref_columns)









import statistics as stat
from typing import Tuple

from .base import Tree

class SpuriousSequence(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree = self.read_tree_file()
        factor = self.factor
        name_and_branch_len, threshold, median = self.identify_spurious_sequence(tree, factor)
        
        counter = 0
        for name, length in name_and_branch_len.items():
            if length >= threshold:
                try:
                    print(f"{name}\t{round(length, 4)}\t{round(threshold, 4)}\t{round(median, 4)}")
                except BrokenPipeError:
                    pass
                counter += 1
        
        # if no terminal branch is longer than the one specified
        # inform the user and print "None"
        if counter == 0:
            print("None")


    def process_args(self, args):
        return dict(tree_file_path=args.tree, factor=args.factor)

    def identify_spurious_sequence(self, tree, factor):
        branch_lengths, name_and_branch_len = self.get_branch_lengths_and_their_names(tree)

        median = stat.median(branch_lengths)

        if factor is None:
            factor = 20

        threshold = median*factor

        return name_and_branch_len, threshold, median
    
    def get_branch_lengths_and_their_names(
        self,
        tree
    ) -> Tuple[list, list]:
        # initialize a list to hold branch lengths and a
        # dictionary with terminal names and the branch
        # lengths that lead up to them.
        branch_lengths = []
        name_and_branch_len = {}
        
        # collect terminal branch lengths
        for terminal in tree.get_terminals():
            branch_lengths.append(terminal.branch_length)
            name_and_branch_len[terminal.name] = terminal.branch_length
        
        # collect internal branch lengths
        for internal in tree.get_nonterminals():
            branch_lengths.append(terminal.branch_length)

        return branch_lengths, name_and_branch_len      




import sys
import getopt
import os.path
import itertools
from scipy.stats import chisquare
from typing import Tuple

from Bio import Phylo
from Bio.Phylo.BaseTree import TreeMixin

from .base import Tree

from ...helpers.files import read_single_column_file_to_list

class PolytomyTest(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        # read in groups
        groups_arr = self.read_in_groups(self.groups)

        # determine groups of groups
        groups_of_groups, outgroup_taxa = self.determine_groups_of_groups(groups_arr)

        # read trees into list
        trees_file_path = read_single_column_file_to_list(self.trees)

        # go through all triplets of all trees and  
        # examine sister relationships among all triplets
        summary = self.loop_through_trees_and_examine_sister_support_among_triplets(
            trees_file_path,
            groups_of_groups,
            outgroup_taxa
        )

        # count triplet and gene support frequencies for different sister relationships
        triplet_group_counts, gene_support_freq = self.get_triplet_and_gene_support_freq_counts(summary)

        # conduct chisquare tests
        triplet_res, gene_support_freq_res = self.chisquare_tests(triplet_group_counts, gene_support_freq)

        # print results
        self.print_gene_support_freq_res(gene_support_freq_res, gene_support_freq, trees_file_path)
        # self.print_triplet_based_res(triplet_res, triplet_group_counts)
    
    def process_args(self, args):
        return dict(trees=args.trees, groups=args.groups)

    def read_in_groups(self, groups) -> list:
        groups_arr = []
        try:
            for line in open(self.groups):
                line = line.strip()
                if not line.startswith("#"):
                    try:
                        line = line.split("\t")
                        temp = []
                        temp.append(line[0])
                        temp.append(line[1].split(";"))
                        temp.append(line[2].split(";"))
                        temp.append(line[3].split(";"))
                        temp.append(line[4].split(";"))
                        groups_arr.append(temp)
                    except IndexError:
                        try:
                            print(f"{self.groups} contains an indexing error.")
                            print("Please format the groups file (-g) as a four column tab-delimited file with column 1 being the name of the test")
                            print("col2: the tip names of one group (; separated)")
                            print("col3: the tip names of a second group (; separated)")
                            print("col4: the tip names of a third group (; separated)")
                            print("col5: the tip names of the outgroup taxa (; separated)")
                            sys.exit()
                        except BrokenPipeError:
                            pass
                        
        except FileNotFoundError:
            try:
                print(f"{self.groups} corresponds to no such file.")
                print("Please check filename and pathing again.")
                sys.exit()
            except BrokenPipeError:
                pass
        return groups_arr

    def loop_through_trees_and_examine_sister_support_among_triplets(
        self,
        trees_file_path: str,
        groups_of_groups: dict,
        outgroup_taxa: list
    ) -> dict:
        """
        go through all trees and all triplets of all trees. For each triplet,
        determine which two taxa are sister to one another
        """
        summary = {}
        # loop through trees
        try:
            #cnt = 0
            for tree_file in trees_file_path:
                #cnt+=1
                #print(f"processing tree {cnt} of {len(trees_file_path)}")
                tree = Phylo.read(tree_file, 'newick')
                # get tip names
                tips = self.get_tip_names_from_tree(tree)

                # examine all triplets and their support for 
                # any sister pairing
                summary = self.examine_all_triplets_and_sister_pairing(
                    tips,
                    tree_file,
                    summary,
                    groups_of_groups,
                    outgroup_taxa
                )
        except FileNotFoundError:
            try:
                print(f"{tree_file} corresponds to no such file.")
                print("Please check file name and pathing")
                sys.exit()
            except BrokenPipeError:
                pass
        
        return summary

    def determine_groups_of_groups(
        self,
        groups_arr: list
    ) -> dict:
        groups_of_groups = {}

        for group in groups_arr:
            temp = []
            for i in range(1, 4):
                temp.append([taxon_name for taxon_name in group[i]])
            groups_of_groups[group[0]] = (temp)

        outgroup_taxa = [taxon_name for taxon_name in group[4]]
        
        return groups_of_groups, outgroup_taxa

    def examine_all_triplets_and_sister_pairing(
        self,
        tips: list,
        tree_file: str,
        summary: dict,
        groups_of_groups: dict,
        outgroup_taxa: list
    ) -> dict:
        """
        evaluate all triplets for sister relationships. Polytomies
        in input trees are accounted for
        """
        # get all combinations of three tips
        identifier = list(groups_of_groups.keys())[0]
        triplet_tips = (list(itertools.product(*groups_of_groups[identifier])))
        for triplet in triplet_tips:
            # obtain tree of the triplet
            tree = self.get_triplet_tree(tips, triplet, tree_file, outgroup_taxa)
            cnt=0
            try:
                for leaf in tree.get_terminals():
                    cnt+=1
            except AttributeError:
                continue
            if tree and cnt==3:
                for _, groups in groups_of_groups.items():
                    # see if there any intersctions between
                    # the triplet and the the group
                    num_groups_represented = self.count_number_of_groups_in_triplet(triplet, groups)
                    
                    # if one taxa is represented from each group, 
                    # use the triplet
                    tip_names = []
                    if num_groups_represented == 3:
                        # get names in triplet and set tree branch lengths to 1
                        tip_names = self.get_tip_names_from_tree(tree)
                        self.set_branch_lengths_in_tree_to_one(tree)
                        # determine sisters and add to sister pair counter
                        summary = self.determine_sisters_and_add_to_counter(
                            tip_names,
                            tree,
                            tree_file,
                            groups,
                            summary
                        )
            else:
                continue

        return summary

    def count_number_of_groups_in_triplet(self, triplet: list, groups: tuple) -> int:
        """
        determine how many groups are represented in a triplet
        """
        num_groups_represented = 0
        for group in groups:
            temp = [value for value in list(triplet) if value in group]
            if len(temp) >= 1:
                num_groups_represented +=1
        return num_groups_represented

    def set_branch_lengths_in_tree_to_one(self, tree: Tree) -> None:
        for term in tree.get_terminals():
            term.branch_length = 1
        for internode in tree.get_nonterminals():
            internode.branch_length = 1

    def check_if_triplet_is_a_polytomy(self, tree: Tree) -> Tree:
        """
        count the number of internal branches. If 1, then the triplet is a polytomy
        """
        num_int=0
        # check if the triplet is a polytomy
        for internal_branch in tree.get_nonterminals():
            num_int+=1

        if num_int == 1:
            return True
        else:
            return False

    def sister_relationship_counter(
        self,
        tree_file: str,
        summary: dict,
        sisters: str
    ) -> dict:
        """
        counter for how many times a particular sister relationship is observed
        """
        # if tree is not in summary, create a key for it
        if tree_file not in summary.keys():
            summary[str(tree_file)]={}
        # if the sister relationship is not in the tree file dict, create a key for it
        if sisters not in summary[str(tree_file)].keys():
            summary[str(tree_file)][sisters] = 1
        else:
            summary[str(tree_file)][sisters] += 1

        return summary

    def get_triplet_tree(
        self,
        tips: list,
        triplet: tuple,
        tree_file: str, 
        outgroup_taxa: list
    ) -> Tree:
        """
        get a tree object of only the triplet of interest
        """
        # determine tips that are not in the triplet of interest
        tips_to_prune = list(set(tips)- set(list(triplet)))
        # determine tips that in the outgroup
        outgroup_present = [value for value in tips if value in outgroup_taxa]
        tree = Phylo.read(tree_file, 'newick')
        
        # root tree on outgroup taxa
        try:
            tree.root_with_outgroup(outgroup_present)

            # prune to a triplet
            tree = self.prune_tree_using_taxa_list(tree, tips_to_prune)

            return tree
        except ValueError:
            tree = False
            return tree

    def determine_sisters_from_triplet(
        self,
        groups: list,
        pair: tuple
    ) -> str:
        """
        determine sister taxa from a triplet
        """
        sisters = sorted([[i for i, lst in enumerate(groups) if pair[0] in lst][0], [i for i, lst in enumerate(groups) if pair[1] in lst][0]])
        sisters = [str(i) for i in sisters] 
        sisters = str("-".join(sisters))

        return sisters

    def determine_sisters_and_add_to_counter(
        self,
        tip_names: list,
        tree: Tree,
        tree_file: str,
        groups: list,
        summary: dict
    ) -> dict:
        """
        determine which pair of taxa are sister to one another
        and add 1 to the counter for the sister pair
        """
        # get pairs from tip names
        pairs = list(itertools.combinations(tip_names, 2))
        for pair in pairs:
            is_polytomy = self.check_if_triplet_is_a_polytomy(tree)
            # if distance between pair is 2 and the triplet is 
            # not a polytomy (i.e., having only 1 internal branch)
            # then report the sisters in the triplet
            if tree.distance(pair[0], pair[1]) == 2 and not is_polytomy:
                # determine which two tips are sisters
                sisters = self.determine_sisters_from_triplet(groups, pair)
                # add to summary dictionary of how many times that sister
                # relationship is observed
                summary = self.sister_relationship_counter(tree_file, summary, sisters)
        return summary

    def get_triplet_and_gene_support_freq_counts(
        self,
        summary: dict
    ) -> Tuple[dict, dict]:
        """
        count how many triplets and genes support the various sister relationships
        """
        # Count the total number of sister pairings
        # for the three possible pairs for triplets
        triplet_group_counts = {
            'g0g1_count' : 0,
            'g0g2_count' : 0, 
            'g1g2_count' : 0
        }

        # Also, keep track of which and how many genes 
        # support each sister pairing
        gene_support_freq = {
            '0-1' : 0,
            '1-2' : 0,
            '0-2' : 0
        }
        for tree in summary:
            # create empty key value pairs in case sister 
            # pairing was never observed
            if '0-1' not in summary[tree].keys():
                summary[tree]['0-1'] = 0 
            if '0-2' not in summary[tree].keys():
                summary[tree]['0-2'] = 0 
            if '1-2' not in summary[tree].keys():
                summary[tree]['1-2'] = 0
            # create a running value of triplets that support each sister pair
            triplet_group_counts['g0g1_count'] += summary[tree]['0-1']
            triplet_group_counts['g0g2_count'] += summary[tree]['0-2']
            triplet_group_counts['g1g2_count'] += summary[tree]['1-2']
            # determine which sister pairing is best supported in a single gene
            # and add one to the corresponding gene support frequency count
            gene_support_freq[max(summary[tree], key=summary[tree].get)] += 1

        return triplet_group_counts, gene_support_freq

    def chisquare_tests(
        self,
        triplet_group_counts: dict,
        gene_support_freq: dict
    ): 
        
        triplet_res = chisquare(
            [
                triplet_group_counts['g0g1_count'],
                triplet_group_counts['g0g2_count'],
                triplet_group_counts['g1g2_count']
            ]
        )
        
        gene_support_freq_res = chisquare(
            [
                gene_support_freq['0-1'],
                gene_support_freq['0-2'],gene_support_freq['1-2']
            ]
        )

        return triplet_res, gene_support_freq_res

    # def print_triplet_based_res(
    #     self,
    #     triplet_res,
    #     triplet_group_counts: dict
    # ) -> None:
    #     """
    #     print results to stdout for user
    #     """
    #     try:
    #         print(f"\nTriplet Results")
    #         print(f"===============")
    #         print(f"chi-squared: {round(triplet_res.statistic, 4)}")
    #         print(f"p-value: {round(triplet_res.pvalue, 6)}")
    #         print(f"total triplets: {sum(triplet_group_counts.values())}")
    #         print(f"0-1: {triplet_group_counts['g0g1_count']}")
    #         print(f"0-2: {triplet_group_counts['g0g2_count']}")
    #         print(f"1-2: {triplet_group_counts['g1g2_count']}")
    #     except BrokenPipeError:
    #         pass

    def print_gene_support_freq_res(
        self,
        gene_support_freq_res,
        gene_support_freq: dict,
        trees_file_path: list
    ) -> None:
        """
        print results to stdout for user
        """
        try:
            print(f"Gene Support Frequency Results")
            print(f"==============================")
            print(f"chi-squared: {round(gene_support_freq_res.statistic, 4)}")
            print(f"p-value: {round(gene_support_freq_res.pvalue, 6)}")
            print(f"total genes: {(gene_support_freq['0-1'] + gene_support_freq['0-2'] + gene_support_freq['1-2'])}")
            print(f"0-1: {gene_support_freq['0-1']}")
            print(f"0-2: {gene_support_freq['0-2']}")
            print(f"1-2: {gene_support_freq['1-2']}")
        except BrokenPipeError:
            pass



import sys
import getopt
import os.path
import statistics as stat

from Bio import Phylo
from Bio.Phylo.BaseTree import TreeMixin
import itertools
import numpy as np

from .base import Tree


class RobinsonFouldsDistance(Tree):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        tree_zero = self.read_tree_file()
        tree_one = self.read_tree1_file()

        # get tree tip names
        tree_zero_tips = self.get_tip_names_from_tree(tree_zero)
        tree_one_tips = self.get_tip_names_from_tree(tree_one)

        # determine shared tips, tips to prune, and prune tips
        shared_tree_tips = self.shared_tips(tree_zero_tips, tree_one_tips)
        tree_zero_tips_to_prune = list(set(tree_zero_tips) - set(shared_tree_tips))
        tree_one_tips_to_prune = list(set(tree_one_tips) - set(shared_tree_tips))
        tree_zero = self.prune_tree_using_taxa_list(tree_zero, tree_zero_tips_to_prune)
        tree_one = self.prune_tree_using_taxa_list(tree_one, tree_one_tips_to_prune)

        tip_for_rooting = ''
        for term in tree_zero.get_terminals():
            tip_for_rooting = term.name
            break
        tree_zero.root_with_outgroup(tip_for_rooting)
        tree_one.root_with_outgroup(tip_for_rooting)

        plain_rf, normalized_rf = self.calculate_robinson_foulds_distance(tree_zero, tree_one)
        
        print(f"{plain_rf}\t{round(normalized_rf, 4)}")


    def process_args(self, args):
        return dict(tree_file_path=args.tree_zero, tree1_file_path=args.tree_one)

    def calculate_robinson_foulds_distance(self, tree_zero, tree_one):
        plain_rf = 0
        plain_rf = self.compare_trees(plain_rf, tree_zero, tree_one)
        plain_rf = self.compare_trees(plain_rf, tree_one, tree_zero)
        # count the number of tips in a phylogeny
        tip_count = tree_zero.count_terminals()
        
        # calculate normalized rf distance
        normalized_rf = (plain_rf/(2*(tip_count-3)))

        return plain_rf, normalized_rf

    def compare_trees(
        self,
        plain_rf: int,
        tree_zero: Tree,
        tree_one: Tree
    ) -> int:
        # loop through tree_zero and find similar clade in tree_one
        for clade_zero in tree_zero.get_nonterminals()[1:]:
            # initialize and populate a list of tip names in tree_zero
            tip_names_zero = self.get_tip_names_from_tree(clade_zero)
            # get common ancestor of tree_zero tip names in tree_one
            clade_one = tree_one.common_ancestor(tip_names_zero)
            # initialize and populate a list of tip names in tree_one
            tip_names_one = self.get_tip_names_from_tree(clade_one)
            # compare the list of tip names
            plain_rf = self.determine_if_clade_differs(
                plain_rf,
                tip_names_zero,
                tip_names_one
            )

        return plain_rf

    def determine_if_clade_differs(
        self,
        plain_rf: int,
        tip_names_zero: list,
        tip_names_one: list
    ) -> int:
        """
        if clade differs, add 1 to plain_rf value
        """
        if set(tip_names_zero) != set(tip_names_one):
            plain_rf +=1
        
        return plain_rf






from Bio.Align import MultipleSeqAlignment

from .base import Alignment

class VariableSites(Alignment):
    def __init__(self, args) -> None:
        super().__init__(**self.process_args(args))

    def run(self):
        alignment, alignment_format = self.get_alignment_and_format()
        var_sites, aln_len, var_sites_per = self.calculate_variable_sites(alignment)
        if (var_sites, aln_len, var_sites_per):
            print(f"{var_sites}\t{aln_len}\t{round(var_sites_per, 4)}")


    def process_args(self, args):
        return dict(alignment_file_path=args.alignment)

    def calculate_variable_sites(self, alignment):
        aln_len = alignment.get_alignment_length()
        
        var_sites = 0
        # count number of variable sites
        for i in range(0, aln_len, int(1)):
            # obtain sequence at position, remove gaps, and make
            # all characters uppercase
            seq_at_position = ""
            seq_at_position += alignment[:, i]
            seq_at_position = seq_at_position.upper().replace("-", "")
            if len(set(seq_at_position)) > 1:
                var_sites += 1
        # calculate percent of variable sites
        var_sites_per = (var_sites / aln_len)*100
        
        return var_sites, aln_len, var_sites_per